"""Implements the xonsh executer"""
from __future__ import print_function, unicode_literals
import re
import os
import types
import inspect
import builtins
from collections import Iterable, Sequence, Mapping

from xonsh import ast
from xonsh.parser import Parser
from xonsh.tools import subproc_toks
from xonsh.built_ins import load_builtins, unload_builtins


class Execer(object):
    """Executes xonsh code in a context."""

    def __init__(self,
                 filename='<xonsh-code>',
                 debug_level=0,
                 parser_args=None,
                 unload=True):
        """Parameters
        ----------
        filename : str, optional
            File we are to execute.
        debug_level : int, optional
            Debugging level to use in lexing and parsing.
        parser_args : dict, optional
            Arguments to pass down to the parser.
        unload : bool, optional
            Whether or not to unload xonsh builtins upon deletion.
        """
        parser_args = parser_args or {}
        self.parser = Parser(**parser_args)
        self.filename = filename
        self.debug_level = debug_level
        self.unload = unload
        self.ctxtransformer = ast.CtxAwareTransformer(self.parser)
        load_builtins(execer=self)

    def __del__(self):
        if self.unload:
            unload_builtins()

    def parse(self, input, ctx, mode='exec'):
        """Parses xonsh code in a context-aware fashion. For context-free
        parsing, please use the Parser class directly.
        """
        if ctx is None:
            ctx = set()
        elif isinstance(ctx, Mapping):
            ctx = set(ctx.keys())

        # Parsing actually happens in a couple of phases. The first is a
        # shortcut for a context-free parser. Normally, all subprocess
        # lines should be wrapped in $(), to indicate that they are a
        # subproc. But that would be super annoying. Unfortnately, Python
        # mode - after indentation - is whitespace agnostic while, using
        # the Python token, subproc mode is whitespace aware. That is to say,
        # in Python mode "ls -l", "ls-l", and "ls - l" all parse to the
        # same AST because whitespace doesn't matter to the minus binary op.
        # However, these phases all have very different meaning in subproc
        # mode. The 'right' way to deal with this is to make the entire
        # grammar whitespace aware, and then ignore all of the whitespace
        # tokens for all of the Python rules. The lazy way implemented here
        # is to parse a line a second time with a $() wrapper if it fails
        # the first time. This is a context-free phase.
        tree = self._parse_ctx_free(input, mode=mode)
        if tree is None:
            return None

        # Now we need to perform context-aware AST transformation. This is
        # because the "ls -l" is valid Python. The only way that we know
        # it is not actually Python is by checking to see if the first token
        # (ls) is part of the execution context. If it isn't, then we will
        # assume that this line is supposed to be a subprocess line, assuming
        # it also is valid as a subprocess line.
        tree = self.ctxtransformer.ctxvisit(tree, input, ctx, mode=mode)
        return tree

    def compile(self, input, mode='exec', glbs=None, locs=None, stacklevel=2):
        """Compiles xonsh code into a Python code object, which may then
        be execed or evaled.
        """
        if glbs is None or locs is None:
            frame = inspect.stack()[stacklevel][0]
            glbs = frame.f_globals if glbs is None else glbs
            locs = frame.f_locals if locs is None else locs
        ctx = set(dir(builtins)) | set(glbs.keys()) | set(locs.keys())
        tree = self.parse(input, ctx, mode=mode)
        if tree is None:
            return None  # handles comment only input
        code = compile(tree, self.filename, mode)
        return code

    def eval(self, input, glbs=None, locs=None, stacklevel=2):
        """Evaluates (and returns) xonsh code."""
        if isinstance(input, types.CodeType):
            code = input
        else:
            code = self.compile(input=input,
                                glbs=glbs,
                                locs=locs,
                                mode='eval',
                                stacklevel=stacklevel)
        if code is None:
            return None  # handles comment only input
        return eval(code, glbs, locs)

    def exec(self, input, mode='exec', glbs=None, locs=None, stacklevel=2):
        """Execute xonsh code."""
        if isinstance(input, types.CodeType):
            code = input
        else:
            code = self.compile(input=input,
                                glbs=glbs,
                                locs=locs,
                                mode=mode,
                                stacklevel=stacklevel)
        if code is None:
            return None  # handles comment only input
        return exec(code, glbs, locs)

    def _find_next_break(self, line, mincol):
        if mincol >= 1:
            line = line[mincol:]
        if ';' not in line:
            return None
        maxcol = None
        self.parser.lexer.input(line)
        for tok in self.parser.lexer:
            if tok.type == 'SEMI':
                maxcol = tok.lexpos + mincol + 1
                break
        return maxcol

    def _parse_ctx_free(self, input, mode='exec'):
        last_error_line = last_error_col = -1
        parsed = False
        original_error = None
        while not parsed:
            try:
                tree = self.parser.parse(input,
                                         filename=self.filename,
                                         mode=mode,
                                         debug_level=self.debug_level)
                parsed = True
            except IndentationError as e:
                if original_error is None:
                    raise e
                else:
                    raise original_error
            except SyntaxError as e:
                if original_error is None:
                    original_error = e
                if (e.loc is None) or (last_error_line == e.loc.lineno and
                                       last_error_col in (e.loc.column + 1,
                                                          e.loc.column)):
                    raise original_error
                last_error_col = e.loc.column
                last_error_line = e.loc.lineno
                idx = last_error_line - 1
                lines = input.splitlines()
                line = lines[idx]
                if input.endswith('\n'):
                    lines.append('')
                if len(line.strip()) == 0:
                    # whitespace only lines are not valid syntax in Python's
                    # interactive mode='single', who knew?! Just ignore them.
                    # this might cause actual sytax errors to have bad line
                    # numbers reported, but should only effect interactive mode
                    del lines[idx]
                    last_error_line = last_error_col = -1
                    input = '\n'.join(lines)
                    continue
                maxcol = self._find_next_break(line, last_error_col)
                sbpline = subproc_toks(line,
                                       returnline=True,
                                       maxcol=maxcol,
                                       lexer=self.parser.lexer)
                if sbpline is None:
                    # subprocess line had no valid tokens, likely because
                    # it only contained a comment.
                    del lines[idx]
                    last_error_line = last_error_col = -1
                    input = '\n'.join(lines)
                    continue
                else:
                    lines[idx] = sbpline
                last_error_col += 3
                input = '\n'.join(lines)
        return tree



"""Import hooks for importing xonsh source files. This module registers
the hooks it defines when it is imported.
"""
from __future__ import print_function, unicode_literals
import os
import sys
import builtins
from importlib.machinery import ModuleSpec
from importlib.abc import MetaPathFinder, SourceLoader

from xonsh.tools import string_types
from xonsh.execer import Execer


class XonshImportHook(MetaPathFinder, SourceLoader):
    """Implements the import hook for xonsh source files."""

    def __init__(self, *args, **kwargs):
        super(XonshImportHook, self).__init__(*args, **kwargs)
        self._filenames = {}
        self._execer = None

    @property
    def execer(self):
        if hasattr(builtins, '__xonsh_execer__'):
            execer = builtins.__xonsh_execer__
            if self._execer is not None:
                self._execer = None
        elif self._execer is None:
            self._execer = execer = Execer(unload=False)
        else:
            execer = self._execer
        return execer

    #
    # MetaPathFinder methods
    #
    def find_spec(self, fullname, path, target=None):
        """Finds the spec for a xonsh module if it exists."""
        dot = '.'
        spec = None
        path = sys.path if path is None else path
        if dot not in fullname and dot not in path:
            path = [dot] + path
        name = fullname.rsplit(dot, 1)[-1]
        fname = name + '.xsh'
        for p in path:
            if not isinstance(p, string_types):
                continue
            if not os.path.isdir(p):
                continue
            if fname not in os.listdir(p):
                continue
            spec = ModuleSpec(fullname, self)
            self._filenames[fullname] = os.path.join(p, fname)
            break
        return spec

    #
    # SourceLoader methods
    #
    def get_filename(self, fullname):
        """Returns the filename for a module's fullname."""
        return self._filenames[fullname]

    def get_data(self, path):
        """Gets the bytes for a path."""
        raise NotImplementedError

    def get_code(self, fullname):
        """Gets the code object for a xonsh file."""
        filename = self._filenames.get(fullname, None)
        if filename is None:
            msg = "xonsh file {0!r} could not be found".format(fullname)
            raise ImportError(msg)
        with open(filename, 'r') as f:
            src = f.read()
        src = src if src.endswith('\n') else src + '\n'
        execer = self.execer
        execer.filename = filename
        ctx = {}  # dummy for modules
        code = execer.compile(src, glbs=ctx, locs=ctx)
        return code


sys.meta_path.append(XonshImportHook())



"""Hooks for pygments syntax highlighting."""
from __future__ import print_function, unicode_literals

from pygments.lexer import inherit, bygroups, using, this
from pygments.token import Name, Generic, Keyword, Text, String
from pygments.lexers.shell import BashLexer
from pygments.lexers.agile import PythonLexer


class XonshSubprocLexer(BashLexer):
    """Lexer for xonsh subproc mode."""

    name = 'Xonsh subprocess lexer'

    tokens = {'root': [(r'`[^`]*?`', String.Backtick), inherit, ]}


ROOT_TOKENS = [(r'\?', Keyword),
               (r'\$\w+', Name.Variable),
               (r'\$\{', Keyword, ('pymode', )),
               (r'\$\(', Keyword, ('subproc', )),
               (r'\$\[', Keyword, ('subproc', )),
               (r'@\(', Keyword, ('pymode', )),
               inherit, ]

PYMODE_TOKENS = [(r'(.+)(\))', bygroups(using(this), Keyword), '#pop'),
                 (r'(.+)(\})', bygroups(using(this), Keyword), '#pop'), ]

SUBPROC_TOKENS = [
    (r'(.+)(\))', bygroups(using(XonshSubprocLexer), Keyword), '#pop'),
    (r'(.+)(\])', bygroups(using(XonshSubprocLexer), Keyword), '#pop'),
]


class XonshLexer(PythonLexer):
    """Xonsh console lexer for pygments."""

    name = 'Xonsh lexer'
    aliases = ['xonsh', 'xsh']
    filenames = ['*.xsh', '*xonshrc']

    tokens = {
        'root': list(ROOT_TOKENS),
        'pymode': PYMODE_TOKENS,
        'subproc': SUBPROC_TOKENS,
    }


class XonshConsoleLexer(PythonLexer):
    """Xonsh console lexer for pygments."""

    name = 'Xonsh console lexer'
    aliases = ['xonshcon']
    filenames = []

    tokens = {
        'root': [(r'^(>>>|\.\.\.) ', Generic.Prompt),
                 (r'\n(>>>|\.\.\.)', Generic.Prompt),
                 (r'\n(?![>.][>.][>.] )([^\n]*)', Generic.Output),
                 (r'\n(?![>.][>.][>.] )(.*?)$', Generic.Output)] + ROOT_TOKENS,
        'pymode': PYMODE_TOKENS,
        'subproc': SUBPROC_TOKENS,
    }

# XonshLexer & XonshSubprocLexer have to refernce each other
XonshSubprocLexer.tokens['root'] = [
    (r'(\$\{)(.*)(\})', bygroups(Keyword, using(XonshLexer), Keyword)),
    (r'(@\()(.+)(\))', bygroups(Keyword, using(XonshLexer), Keyword)),
] + XonshSubprocLexer.tokens['root']



"""The main xonsh script."""
import os
import sys
import shlex
import signal
import builtins
import subprocess
from argparse import ArgumentParser, Namespace

from xonsh.shell import Shell

parser = ArgumentParser(description='xonsh')
parser.add_argument('-c',
                    help="Run a single command and exit",
                    dest='command',
                    required=False,
                    default=None)
parser.add_argument('--no-rc',
                    help="Do not load the .xonshrc file",
                    dest='norc',
                    action='store_true',
                    default=False)
parser.add_argument('file',
                    metavar='script-file',
                    help='If present, execute the script in script-file'
                         ' and exit',
                    nargs='?',
                    default=None)
parser.add_argument('args',
                    metavar='args',
                    help='Additional arguments to the script specified'
                         ' by script-file',
                    nargs='*',
                    default=[])


def main(argv=None):
    """Main entry point for xonsh cli."""
    args = parser.parse_args()
    shell = Shell() if not args.norc else Shell(ctx={})
    from xonsh import imphooks
    env = builtins.__xonsh_env__
    env['XONSH_INTERACTIVE'] = False
    if args.command is not None:
        # run a single command and exit
        shell.default(args.command)
    elif args.file is not None:
        # run a script contained in a file
        if os.path.isfile(args.file):
            with open(args.file) as f:
                code = f.read()
            code = code if code.endswith('\n') else code + '\n'
            env['ARGS'] = [args.file] + args.args
            code = shell.execer.compile(code, mode='exec', glbs=shell.ctx)
            shell.execer.exec(code, mode='exec', glbs=shell.ctx)
        else:
            print('xonsh: {0}: No such file or directory.'.format(args.file))
    elif not sys.stdin.isatty():
        # run a script given on stdin
        code = sys.stdin.read()
        code = code if code.endswith('\n') else code + '\n'
        code = shell.execer.compile(code, mode='exec', glbs=shell.ctx)
        shell.execer.exec(code, mode='exec', glbs=shell.ctx)
    else:
        # otherwise, enter the shell
        env['XONSH_INTERACTIVE'] = True
        signal.signal(signal.SIGTSTP, signal.SIG_IGN)
        shell.cmdloop()


if __name__ == '__main__':
    main()



"""
Lexer for xonsh code, written using a hybrid of ``tokenize`` and PLY
"""
from __future__ import print_function, unicode_literals
import re
import sys
import tokenize

from io import BytesIO
from keyword import kwlist

from ply import lex
from ply.lex import TOKEN, LexToken


token_map = {}
"""
Mapping from ``tokenize`` tokens (or token types) to PLY token types.  If a
simple one-to-one mapping from ``tokenize`` to PLY exists, the lexer will look
it up here and generate a single PLY token of the given type.  Otherwise, it
will fall back to handling that token using one of the handlers in
``special_handlers``.
"""

# keywords
for kw in kwlist:
    token_map[(tokenize.NAME, kw)] = kw.upper()

# operators
_op_map = {
        # punctuation
        ',': 'COMMA', '.': 'PERIOD', ';': 'SEMI', ':': 'COLON',
        '...': 'ELLIPSIS',
        # basic operators
        '+': 'PLUS', '-': 'MINUS', '*': 'TIMES', '/': 'DIVIDE',
        '//': 'DOUBLEDIV', '%': 'MOD', '**': 'POW', '|': 'PIPE',
        '&': 'AMPERSAND', '~': 'TILDE', '^': 'XOR', '<<': 'LSHIFT',
        '>>': 'RSHIFT', '<': 'LT', '<=': 'LE', '>': 'GT', '>=': 'GE',
        '==': 'EQ', '!=': 'NE', '->': 'RARROW',
        # assignment operators
        '=': 'EQUALS', '+=': 'PLUSEQUAL', '-=': 'MINUSEQUAL',
        '*=': 'TIMESEQUAL', '/=': 'DIVEQUAL', '%=': 'MODEQUAL',
        '**=': 'POWEQUAL', '<<=': 'LSHIFTEQUAL', '>>=': 'RSHIFTEQUAL',
        '&=': 'AMPERSANDEQUAL', '^=': 'XOREQUAL', '|=': 'PIPEEQUAL',
        '//=': 'DOUBLEDIVEQUAL',
}
for (op, type) in _op_map.items():
    token_map[(tokenize.OP, op)] = type

token_map[tokenize.NAME] = 'NAME'
token_map[tokenize.NUMBER] = 'NUMBER'
token_map[tokenize.STRING] = 'STRING'
token_map[tokenize.NEWLINE] = 'NEWLINE'
token_map[tokenize.INDENT] = 'INDENT'
token_map[tokenize.DEDENT] = 'DEDENT'


def handle_dollar(state, token, stream):
    """
    Function for generating PLY tokens associated with ``$``.
    """
    n = next(stream, None)

    if n is None:
        m = "missing token after $"
        yield _new_token("ERRORTOKEN", m, token.start)
    elif n.start != token.end:
        m = "unexpected whitespace after $"
        yield _new_token("ERRORTOKEN", m, token.start)
    elif n.type == tokenize.NAME:
        state['last'] = n
        yield _new_token('DOLLAR_NAME', '$' + n.string, token.start)
    elif n.type == tokenize.OP and n.string == '(':
        state['pymode'].append((False, '$(', ')', token.start))
        state['last'] = n
        yield _new_token('DOLLAR_LPAREN', '$(', token.start)
    elif n.type == tokenize.OP and n.string == '[':
        state['pymode'].append((False, '$[', ']', token.start))
        state['last'] = n
        yield _new_token('DOLLAR_LBRACKET', '$[', token.start)
    elif n.type == tokenize.OP and n.string == '{':
        state['pymode'].append((True, '${', '}', token.start))
        state['last'] = n
        yield _new_token('DOLLAR_LBRACE', '${', token.start)
    else:
        e = 'expected NAME, (, [, or {{ after $, but got {0}'
        m = e.format(n)
        yield _new_token("ERRORTOKEN", m, token.start)


def handle_at(state, token, stream):
    """
    Function for generating PLY tokens associated with ``@``.
    """
    n = next(stream, None)

    if n is None:
        state['last'] = token
        m = "missing token after @"
        yield _new_token("ERRORTOKEN", m, token.start)
    elif n.type == tokenize.OP and n.string == '(' and \
            n.start == token.end:
        state['pymode'].append((True, '@(', ')', token.start))
        state['last'] = n
        yield _new_token('AT_LPAREN', '@(', token.start)
    else:
        state['last'] = token
        yield _new_token('AT', '@', token.start)
        yield from handle_token(state, n, stream)


def handle_question(state, token, stream):
    """
    Function for generating PLY tokens for help and superhelp
    """
    n = next(stream, None)

    if n is not None and n.type == tokenize.ERRORTOKEN and \
            n.string == '?' and n.start == token.end:
        state['last'] = n
        yield _new_token('DOUBLE_QUESTION', '??', token.start)
    else:
        state['last'] = token
        yield _new_token('QUESTION', '?', token.start)
        if n is not None:
            yield from handle_token(state, n, stream)


def handle_backtick(state, token, stream):
    """
    Function for generating PLY tokens representing regex globs.
    """
    n = next(stream, None)

    found_match = False
    sofar = '`'
    while n is not None:
        sofar += n.string
        if n.type == tokenize.ERRORTOKEN and n.string == '`':
            found_match = True
            break
        elif n.type == tokenize.NEWLINE or n.type == tokenize.NL:
            break
        n = next(stream, None)
    if found_match:
        state['last'] = n
        yield _new_token('REGEXPATH', sofar, token.start)
    else:
        state['last'] = token
        e = "Could not find matching backtick for regex on line {0}"
        m = e.format(token.start[0])
        yield _new_token("ERRORTOKEN", m, token.start)


def handle_lparen(state, token, stream):
    """
    Function for handling ``(``
    """
    state['pymode'].append((True, '(', ')', token.start))
    state['last'] = token
    yield _new_token('LPAREN', '(', token.start)


def handle_lbrace(state, token, stream):
    """
    Function for handling ``{``
    """
    state['pymode'].append((True, '{', '}', token.start))
    state['last'] = token
    yield _new_token('LBRACE', '{', token.start)


def handle_lbracket(state, token, stream):
    """
    Function for handling ``[``
    """
    state['pymode'].append((True, '[', ']', token.start))
    state['last'] = token
    yield _new_token('LBRACKET', '[', token.start)


def _end_delimiter(state, token):
    py = state['pymode']
    s = token.string
    l, c = token.start
    if len(py) > 1:
        mode, orig, match, pos = py.pop()
        if s != match:
            e = '"{}" at {} ends "{}" at {} (expected "{}")'
            return e.format(s, (l, c), orig, pos, match)
    else:
        return 'Unmatched "{}" at line {}, column {}'.format(s, l, c)


def handle_rparen(state, token, stream):
    """
    Function for handling ``)``
    """
    e = _end_delimiter(state, token)
    if e is None:
        state['last'] = token
        yield _new_token('RPAREN', ')', token.start)
    else:
        yield _new_token('ERRORTOKEN', e, token.start)


def handle_rbrace(state, token, stream):
    """
    Function for handling ``}``
    """
    e = _end_delimiter(state, token)
    if e is None:
        state['last'] = token
        yield _new_token('RBRACE', '}', token.start)
    else:
        yield _new_token('ERRORTOKEN', e, token.start)


def handle_rbracket(state, token, stream):
    """
    Function for handling ``]``
    """
    e = _end_delimiter(state, token)
    if e is None:
        state['last'] = token
        yield _new_token('RBRACKET', ']', token.start)
    else:
        yield _new_token('ERRORTOKEN', e, token.start)


def handle_error_space(state, token, stream):
    """
    Function for handling special whitespace characters is subprocess mode
    """
    if not state['pymode'][-1][0]:
        state['last'] = token
        yield _new_token('WS', token.string, token.start)
    else:
        yield from []


def handle_ignore(state, token, stream):
    """
    Function for handling tokens that should be ignored
    """
    yield from []


special_handlers = {
    tokenize.NL: handle_ignore,
    tokenize.COMMENT: handle_ignore,
    tokenize.ENCODING: handle_ignore,
    tokenize.ENDMARKER: handle_ignore,
    (tokenize.OP, '@'): handle_at,
    (tokenize.OP, '('): handle_lparen,
    (tokenize.OP, ')'): handle_rparen,
    (tokenize.OP, '{'): handle_lbrace,
    (tokenize.OP, '}'): handle_rbrace,
    (tokenize.OP, '['): handle_lbracket,
    (tokenize.OP, ']'): handle_rbracket,
    (tokenize.ERRORTOKEN, '$'): handle_dollar,
    (tokenize.ERRORTOKEN, '`'): handle_backtick,
    (tokenize.ERRORTOKEN, '?'): handle_question,
    (tokenize.ERRORTOKEN, ' '): handle_error_space,
}
"""
Mapping from ``tokenize`` tokens (or token types) to the proper function for
generating PLY tokens from them.  In addition to yielding PLY tokens, these
functions may manipulate the Lexer's state.
"""


def handle_token(state, token, stream):
    """
    General-purpose token handler.  Makes use of ``token_map`` or
    ``special_map`` to yield one or more PLY tokens from the given input.

    Parameters
    ----------

    state :
        The current state of the lexer, including information about whether
        we are in Python mode or subprocess mode, which changes the lexer's
        behavior
    token :
        The token (from ``tokenize``) currently under consideration
    stream :
        A generator from which more tokens can be grabbed if necessary
    """
    typ = token.type
    st = token.string
    pymode = state['pymode'][-1][0]
    if not pymode:
        if state['last'] is not None and state['last'].end != token.start:
            cur = token.start
            old = state['last'].end
            if cur[0] == old[0] and cur[1] > old[1]:
                yield _new_token('WS', token.line[old[1]:cur[1]], old)
    if typ == tokenize.NAME and not pymode:
        state['last'] = token
        yield _new_token('NAME', st, token.start)
    elif (typ, st) in token_map:
        state['last'] = token
        yield _new_token(token_map[(typ, st)], st, token.start)
    elif typ in token_map:
        state['last'] = token
        yield _new_token(token_map[typ], st, token.start)
    elif (typ, st) in special_handlers:
        yield from special_handlers[(typ, st)](state, token, stream)
    elif typ in special_handlers:
        yield from special_handlers[typ](state, token, stream)
    else:
        m = "Unexpected token: {0}".format(token)
        yield _new_token("ERRORTOKEN", m, token.start)


def get_tokens(s):
    """
    Given a string containing xonsh code, generates a stream of relevant PLY
    tokens using ``handle_token``.
    """
    tokstream = tokenize.tokenize(BytesIO(s.encode('utf-8')).readline)
    state = {'indents': [0], 'pymode': [(True, '', '', (0, 0))], 'last': None}
    while True:
        try:
            token = next(tokstream)
            yield from handle_token(state, token, tokstream)
        except StopIteration:
            if len(state['pymode']) > 1:
                pm, o, m, p = state['pymode'][-1]
                l, c = p
                e = 'Unmatched "{}" at line {}, column {}'
                yield _new_token('ERRORTOKEN', e.format(o, l, c), (0, 0))
            break
        except tokenize.TokenError as e:
            # this is recoverable in single-line mode (from the shell)
            # (e.g., EOF while scanning string literal)
            yield _new_token('ERRORTOKEN', e.args[0], (0, 0))
            break
        except IndentationError as e:
            # this is never recoverable
            yield _new_token('ERRORTOKEN', e, (0, 0))
            break


# synthesize a new PLY token
def _new_token(type, value, pos):
    o = LexToken()
    o.type = type
    o.value = value
    o.lineno, o.lexpos = pos
    return o


class Lexer(object):
    """Implements a lexer for the xonsh language."""

    def __init__(self):
        """
        Attributes
        ----------
        fname : str
            Filename
        last : token
            The last token seen.
        lineno : int
            The last line number seen.

        """
        self.fname = ''
        self.last = None

    def build(self, **kwargs):
        """Part of the PLY lexer API."""
        pass

    def reset(self):
        pass

    def input(self, s):
        """Calls the lexer on the string s."""
        self.token_stream = get_tokens(s)

    def token(self):
        """Retrieves the next token."""
        self.last = next(self.token_stream, None)
        return self.last

    def __iter__(self):
        t = self.token()
        while t is not None:
            yield t
            t = self.token()

    #
    # All the tokens recognized by the lexer
    #
    tokens = tuple(token_map.values()) + (
        'WS',                    # whitespace in subprocess mode
        'REGEXPATH',             # regex escaped with backticks
        'LPAREN', 'RPAREN',      # ( )
        'LBRACKET', 'RBRACKET',  # [ ]
        'LBRACE', 'RBRACE',      # { }
        'AT',                    # @
        'QUESTION',              # ?
        'DOUBLE_QUESTION',       # ??
        'AT_LPAREN',             # @(
        'DOLLAR_NAME',           # $NAME
        'DOLLAR_LPAREN',         # $(
        'DOLLAR_LBRACE',         # ${
        'DOLLAR_LBRACKET',       # $[
        )



"""A (tab-)completer for xonsh."""
from __future__ import print_function, unicode_literals
import os
import re
import builtins
import subprocess

from xonsh.built_ins import iglobpath
from xonsh.tools import subexpr_from_unbalanced

RE_DASHF = re.compile(r'-F\s+(\w+)')
RE_ATTR = re.compile(r'(\S+(\..+)*)\.(\w*)$')

XONSH_TOKENS = {
    'and ', 'as ', 'assert ', 'break', 'class ', 'continue', 'def ', 'del ',
    'elif ', 'else', 'except ', 'finally:', 'for ', 'from ', 'global ',
    'import ', 'if ', 'in ', 'is ', 'lambda ', 'nonlocal ', 'not ', 'or ',
    'pass', 'raise ', 'return ', 'try:', 'while ', 'with ', 'yield ', '+', '-',
    '/', '//', '%', '**', '|', '&', '~', '^', '>>', '<<', '<', '<=', '>', '>=',
    '==', '!=', '->', '=', '+=', '-=', '*=', '/=', '%=', '**=', '>>=', '<<=',
    '&=', '^=', '|=', '//=', ',', ';', ':', '?', '??', '$(', '${', '$[', '..',
    '...'
}

BASH_COMPLETE_SCRIPT = """source {filename}
COMP_WORDS=({line})
COMP_LINE="{line}"
COMP_POINT=${{#COMP_LINE}}
COMP_COUNT={end}
COMP_CWORD={n}
{func} {cmd} {prefix} {prev}
for ((i=0;i<${{#COMPREPLY[*]}};i++)) do echo ${{COMPREPLY[i]}}; done
"""


class Completer(object):
    """This provides a list of optional completions for the xonsh shell."""

    def __init__(self):
        # initialize command cache
        self._path_checksum = None
        self._alias_checksum = None
        self._path_mtime = -1
        self._cmds_cache = frozenset()
        try:
            # FIXME this could be threaded for faster startup times
            self._load_bash_complete_funcs()
            # or we could make this lazy
            self._load_bash_complete_files()
            self.have_bash = True
        except subprocess.CalledProcessError:
            self.have_bash = False

    def complete(self, prefix, line, begidx, endidx, ctx=None):
        """Complete the string s, given a possible execution context.

        Parameters
        ----------
        prefix : str
            The string to match
        line : str
            The line that prefix appears on.
        begidx : int
            The index in line that prefix starts on.
        endidx : int
            The index in line that prefix ends on.
        ctx : Iterable of str (ie dict, set, etc), optional
            Names in the current execution context.

        Returns
        -------
        rtn : list of str
            Possible completions of prefix, sorted alphabetically.
        """
        space = ' '  # intern some strings for faster appending
        slash = '/'
        dot = '.'
        ctx = ctx or {}
        cmd = line.split(' ', 1)[0]
        if begidx == 0:
            # the first thing we're typing; could be python or subprocess, so
            # anything goes.
            rtn = self.cmd_complete(prefix)
        elif cmd in self.bash_complete_funcs:
            rtn = set()
            for s in self.bash_complete(prefix, line, begidx, endidx):
                if os.path.isdir(s.rstrip()):
                    s = s.rstrip() + slash
                rtn.add(s)
            if len(rtn) == 0:
                rtn = self.path_complete(prefix)
            return sorted(rtn)
        elif prefix.startswith('${') or prefix.startswith('@('):
            # python mode explicitly
            rtn = set()
        elif cmd not in ctx:
            if cmd in self._all_commands():
                # subproc mode; do path completions
                return sorted(self.path_complete(prefix))
            else:
                # if we're here, could be anything
                rtn = set()
        else:
            # if we're here, we're not a command, but could be anything else
            rtn = set()
        rtn |= {s for s in XONSH_TOKENS if s.startswith(prefix)}
        if ctx is not None:
            if dot in prefix:
                rtn |= self.attr_complete(prefix, ctx)
            else:
                rtn |= {s for s in ctx if s.startswith(prefix)}
        rtn |= {s for s in dir(builtins) if s.startswith(prefix)}
        rtn |= {s + space for s in builtins.aliases if s.startswith(prefix)}
        rtn |= self.path_complete(prefix)
        return sorted(rtn)

    def _add_env(self, paths, prefix):
        if prefix.startswith('$'):
            env = builtins.__xonsh_env__
            key = prefix[1:]
            paths.update({'$' + k for k in env if k.startswith(key)})

    def _add_dots(self, paths, prefix):
        if prefix in {'', '.'}:
            paths.update({'./', '../'})
        if prefix == '..':
            paths.add('../')

    def cmd_complete(self, cmd):
        """Completes a command name based on what is on the $PATH"""
        space = ' '
        return {s + space for s in self._all_commands() if s.startswith(cmd)}

    def path_complete(self, prefix):
        """Completes based on a path name."""
        space = ' '  # intern some strings for faster appending
        slash = '/'
        tilde = '~'
        paths = set()
        if prefix.startswith("'") or prefix.startswith('"'):
            prefix = prefix[1:]
        for s in iglobpath(prefix + '*'):
            if space in s:
                s = repr(s + (slash if os.path.isdir(s) else ''))
            else:
                s = s + (slash if os.path.isdir(s) else space)
            paths.add(s)
        if tilde in prefix:
            home = os.path.expanduser(tilde)
            paths = {s.replace(home, tilde) for s in paths}
        self._add_env(paths, prefix)
        self._add_dots(paths, prefix)
        return paths

    def bash_complete(self, prefix, line, begidx, endidx):
        """Attempts BASH completion."""
        splt = line.split()
        cmd = splt[0]
        func = self.bash_complete_funcs.get(cmd, None)
        fnme = self.bash_complete_files.get(cmd, None)
        if func is None or fnme is None:
            return set()
        idx = n = 0
        for n, tok in enumerate(splt):
            if tok == prefix:
                idx = line.find(prefix, idx)
                if idx >= begidx:
                    break
            prev = tok
        if len(prefix) == 0:
            prefix = '""'
            n += 1
        script = BASH_COMPLETE_SCRIPT.format(filename=fnme,
                                             line=line,
                                             n=n,
                                             func=func,
                                             cmd=cmd,
                                             end=endidx + 1,
                                             prefix=prefix,
                                             prev=prev)
        out = subprocess.check_output(['bash'],
                                      input=script,
                                      universal_newlines=True,
                                      stderr=subprocess.PIPE)
        space = ' '
        rtn = {s + space if s[-1:].isalnum() else s for s in out.splitlines()}
        return rtn

    def _source_completions(self):
        srcs = []
        for f in builtins.__xonsh_env__.get('BASH_COMPLETIONS', ()):
            if os.path.isfile(f):
                srcs.append('source ' + f)
        return srcs

    def _load_bash_complete_funcs(self):
        self.bash_complete_funcs = bcf = {}
        inp = self._source_completions()
        if len(inp) == 0:
            return
        inp.append('complete -p\n')
        out = subprocess.check_output(['bash'], input='\n'.join(inp),
                                      universal_newlines=True)
        for line in out.splitlines():
            head, cmd = line.rsplit(' ', 1)
            if len(cmd) == 0 or cmd == 'cd':
                continue
            m = RE_DASHF.search(head)
            if m is None:
                continue
            bcf[cmd] = m.group(1)

    def _load_bash_complete_files(self):
        inp = self._source_completions()
        if len(inp) == 0:
            self.bash_complete_files = {}
            return
        inp.append('shopt -s extdebug')
        declare_f = 'declare -F '
        inp += [declare_f + f for f in self.bash_complete_funcs.values()]
        inp.append('shopt -u extdebug\n')
        out = subprocess.check_output(['bash'], input='\n'.join(inp),
                                      universal_newlines=True)
        func_files = {}
        for line in out.splitlines():
            parts = line.split()
            func_files[parts[0]] = parts[-1]
        self.bash_complete_files = {
            cmd: func_files[func]
            for cmd, func in self.bash_complete_funcs.items()
            if func in func_files
        }

    def attr_complete(self, prefix, ctx):
        """Complete attributes of an object."""
        attrs = set()
        m = RE_ATTR.match(prefix)
        if m is None:
            return attrs
        expr, attr = m.group(1, 3)
        expr = subexpr_from_unbalanced(expr, '(', ')')
        expr = subexpr_from_unbalanced(expr, '[', ']')
        expr = subexpr_from_unbalanced(expr, '{', '}')
        try:
            val = builtins.evalx(expr, glbs=ctx)
        except:  # pylint:disable=bare-except
            try:
                val = builtins.evalx(expr, glbs=builtins.__dict__)
            except:  # pylint:disable=bare-except
                return attrs  # anything could have gone wrong!
        opts = dir(val)
        if len(attr) == 0:
            opts = [o for o in opts if not o.startswith('_')]
        else:
            opts = [o for o in opts if o.startswith(attr)]
        prelen = len(prefix)
        for opt in opts:
            a = getattr(val, opt)
            rpl = opt + '(' if callable(a) else opt
            # note that prefix[:prelen-len(attr)] != prefix[:-len(attr)]
            # when len(attr) == 0.
            comp = prefix[:prelen-len(attr)] + rpl
            attrs.add(comp)
        return attrs

    def _all_commands(self):
        path = builtins.__xonsh_env__.get('PATH', [])
        # did PATH change?
        path_hash = hash(tuple(path))
        cache_valid = path_hash == self._path_checksum
        self._path_checksum = path_hash
        # did aliases change?
        al_hash = hash(tuple(sorted(builtins.aliases.keys())))
        self._alias_checksum = al_hash
        cache_valid = cache_valid and al_hash == self._alias_checksum
        pm = self._path_mtime
        # did the contents of any directory in PATH change?
        for d in filter(os.path.isdir, path):
            m = os.stat(d).st_mtime
            if m > pm:
                pm = m
                cache_valid = False
        self._path_mtime = pm
        if cache_valid:
            return self._cmds_cache
        allcmds = set()
        for d in filter(os.path.isdir, path):
            allcmds |= set(os.listdir(d))
        allcmds |= set(builtins.aliases.keys())
        self._cmds_cache = frozenset(allcmds)
        return self._cmds_cache



"""The xonsh built-ins. Note that this module is named 'built_ins' so as
not to be confused with the special Python builtins module.
"""
import os
import re
import sys
import time
import shlex
import signal
import locale
import builtins
import subprocess
from io import TextIOWrapper, StringIO
from glob import glob, iglob
from subprocess import Popen, PIPE
from contextlib import contextmanager
from collections import Sequence, MutableMapping, Iterable, namedtuple, \
    MutableSequence, MutableSet

from xonsh.tools import string_types, redirect_stdout, redirect_stderr
from xonsh.tools import suggest_commands, XonshError
from xonsh.inspectors import Inspector
from xonsh.environ import default_env
from xonsh.aliases import DEFAULT_ALIASES, bash_aliases
from xonsh.jobs import print_one_job, get_next_job_number, wait_for_active_job
from xonsh.jobs import ProcProxy

ENV = None
BUILTINS_LOADED = False
INSPECTOR = Inspector()
LOCALE_CATS = {
    'LC_CTYPE': locale.LC_CTYPE,
    'LC_MESSAGES': locale.LC_MESSAGES,
    'LC_COLLATE': locale.LC_COLLATE,
    'LC_NUMERIC': locale.LC_NUMERIC,
    'LC_MONETARY': locale.LC_MONETARY,
    'LC_TIME': locale.LC_TIME
}


class Env(MutableMapping):
    """A xonsh environment, whose variables have limited typing
    (unlike BASH). Most variables are, by default, strings (like BASH).
    However, the following rules also apply based on variable-name:

    * PATH: any variable whose name ends in PATH is a list of strings.
    * XONSH_HISTORY_SIZE: this variable is an int.
    * LC_* (locale categories): locale catergory names get/set the Python
      locale via locale.getlocale() and locale.setlocale() functions.

    An Env instance may be converted to an untyped version suitable for
    use in a subprocess.
    """

    _arg_regex = re.compile(r'ARG(\d+)')

    def __init__(self, *args, **kwargs):
        """If no initial environment is given, os.environ is used."""
        self._d = {}
        if len(args) == 0 and len(kwargs) == 0:
            args = (os.environ, )
        for key, val in dict(*args, **kwargs).items():
            self[key] = val
        self._detyped = None
        self._orig_env = None

    def detype(self):
        if self._detyped is not None:
            return self._detyped
        ctx = {}
        for key, val in self._d.items():
            if callable(val) or isinstance(val, MutableMapping):
                continue
            if not isinstance(key, string_types):
                key = str(key)
            if 'PATH' in key:
                val = os.pathsep.join(val)
            elif not isinstance(val, string_types):
                val = str(val)
            ctx[key] = val
        self._detyped = ctx
        return ctx

    def replace_env(self):
        """Replaces the contents of os.environ with a detyped version
        of the xonsh environement.
        """
        if self._orig_env is None:
            self._orig_env = dict(os.environ)
        os.environ.clear()
        os.environ.update(self.detype())

    def undo_replace_env(self):
        """Replaces the contents of os.environ with a detyped version
        of the xonsh environement.
        """
        if self._orig_env is not None:
            os.environ.clear()
            os.environ.update(self._orig_env)
            self._orig_env = None

    #
    # Mutable mapping interface
    #

    def __getitem__(self, key):
        m = self._arg_regex.match(key)
        if (m is not None) and (key not in self._d) and ('ARGS' in self._d):
            args = self._d['ARGS']
            ix = int(m.group(1))
            if ix >= len(args):
                e = "Not enough arguments given to access ARG{0}."
                raise IndexError(e.format(ix))
            return self._d['ARGS'][ix]
        val = self._d[key]
        if isinstance(val, (MutableSet, MutableSequence, MutableMapping)):
            self._detyped = None
        return self._d[key]

    def __setitem__(self, key, val):
        if isinstance(key, string_types) and 'PATH' in key:
            val = val.split(os.pathsep) if isinstance(val, string_types) \
                  else val
        elif key == 'XONSH_HISTORY_SIZE' and not isinstance(val, int):
            val = int(val)
        elif key in LOCALE_CATS:
            locale.setlocale(LOCALE_CATS[key], val)
            val = locale.setlocale(LOCALE_CATS[key])
        self._d[key] = val
        self._detyped = None

    def __delitem__(self, key):
        del self._d[key]
        self._detyped = None

    def __iter__(self):
        yield from self._d

    def __len__(self):
        return len(self._d)

    def __str__(self):
        return str(self._d)

    def __repr__(self):
        return '{0}.{1}({2})'.format(self.__class__.__module__,
                                     self.__class__.__name__, self._d)


class Aliases(MutableMapping):
    """Represents a location to hold and look up aliases."""

    def __init__(self, *args, **kwargs):
        self._raw = {}
        self.update(*args, **kwargs)

    def get(self, key, default=None):
        """Returns the (possibly modified) value. If the key is not present,
        then `default` is returned.
        If the value is callable, it is returned without modification. If it
        is an iterable of strings it will be evaluated recursively to expand
        other aliases, resulting in a new list or a "partially applied"
        callable.
        """
        val = self._raw.get(key)
        if val is None:
            return default
        elif isinstance(val, Iterable) or callable(val):
            return self.eval_alias(val, seen_tokens={key})
        else:
            msg = 'alias of {!r} has an inappropriate type: {!r}'
            raise TypeError(msg.format(key, val))

    def eval_alias(self, value, seen_tokens, acc_args=[]):
        """
        "Evaluates" the alias `value`, by recursively looking up the leftmost
        token and "expanding" if it's also an alias.

        A value like ["cmd", "arg"] might transform like this:
        > ["cmd", "arg"] -> ["ls", "-al", "arg"] -> callable()
        where `cmd=ls -al` and `ls` is an alias with its value being a
        callable.  The resulting callable will be "partially applied" with
        ["-al", "arg"].
        """
        # Beware of mutability: default values for keyword args are evaluated
        # only once.
        if callable(value):
            if acc_args:  # Partial application
                return lambda args, stdin=None: value(acc_args + args,
                                                      stdin=stdin)
            else:
                return value
        else:
            token, *rest = value
            if token in seen_tokens or token not in self._raw:
                # ^ Making sure things like `egrep=egrep --color=auto` works,
                # and that `l` evals to `ls --color=auto -CF` if `l=ls -CF`
                # and `ls=ls --color=auto`
                return value + acc_args
            else:
                return self.eval_alias(self._raw[token],
                                       seen_tokens | {token},
                                       rest + acc_args)

    #
    # Mutable mapping interface
    #

    def __getitem__(self, key):
        return self._raw[key]

    def __setitem__(self, key, val):
        if isinstance(val, string_types):
            self._raw[key] = shlex.split(val)
        else:
            self._raw[key] = val

    def __delitem__(self, key):
        del self._raw[key]

    def update(self, *args, **kwargs):
        for key, val in dict(*args, **kwargs).items():
            self[key] = val

    def __iter__(self):
        yield from self._raw

    def __len__(self):
        return len(self._raw)

    def __str__(self):
        return str(self._raw)

    def __repr__(self):
        return '{0}.{1}({2})'.format(self.__class__.__module__,
                                     self.__class__.__name__, self._raw)


def helper(x, name=''):
    """Prints help about, and then returns that variable."""
    INSPECTOR.pinfo(x, oname=name, detail_level=0)
    return x


def superhelper(x, name=''):
    """Prints help about, and then returns that variable."""
    INSPECTOR.pinfo(x, oname=name, detail_level=1)
    return x


def expand_path(s):
    """Takes a string path and expands ~ to home and environment vars."""
    global ENV
    if ENV is not None:
        ENV.replace_env()
    return os.path.expanduser(os.path.expandvars(s))


def reglob(path, parts=None, i=None):
    """Regular expression-based globbing."""
    if parts is None:
        parts = path.split(os.sep)
        d = os.sep if path.startswith(os.sep) else '.'
        return reglob(d, parts=parts, i=0)
    base = subdir = path
    if i == 0:
        if base == '.':
            base = ''
        elif base == '/' and len(parts) > 1:
            i += 1
    regex = re.compile(os.path.join(base, parts[i]))
    files = os.listdir(subdir)
    files.sort()
    paths = []
    i1 = i + 1
    if i1 == len(parts):
        for f in files:
            p = os.path.join(base, f)
            if regex.match(p) is not None:
                paths.append(p)
    else:
        for f in files:
            p = os.path.join(base, f)
            if regex.match(p) is None or not os.path.isdir(p):
                continue
            paths += reglob(p, parts=parts, i=i1)
    return paths


def regexpath(s):
    """Takes a regular expression string and returns a list of file
    paths that match the regex.
    """
    s = expand_path(s)
    return reglob(s)


def globpath(s):
    """Simple wrapper around glob that also expands home and env vars."""
    s = expand_path(s)
    o = glob(s)
    return o if len(o) != 0 else [s]


def iglobpath(s):
    """Simple wrapper around iglob that also expands home and env vars."""
    s = expand_path(s)
    return iglob(s)


WRITER_MODES = {'>': 'w', '>>': 'a'}


def _run_callable_subproc(alias, args,
                          captured=True,
                          prev_proc=None,
                          stdout=None):
    """Helper for running callables as a subprocess."""
    # compute stdin for callable
    if prev_proc is None:
        stdin = None
    elif isinstance(prev_proc, ProcProxy):
        stdin = prev_proc.stdout
    else:
        stdin = StringIO(prev_proc.communicate()[0].decode(), None)
        stdin.seek(0)
        stdin, _ = stdin.read(), stdin.close()
    # Redirect the output streams temporarily. merge with possible
    # return values from alias function.
    if stdout is PIPE:
        # handles captured mode
        new_stdout, new_stderr = StringIO(), StringIO()
        with redirect_stdout(new_stdout), redirect_stderr(new_stderr):
            rtn = alias(args, stdin=stdin)
        proxy_stdout = new_stdout.getvalue()
        proxy_stderr = new_stderr.getvalue()
        if isinstance(rtn, str):
            proxy_stdout += rtn
        elif isinstance(rtn, Sequence):
            if rtn[0]:  # not None nor ''
                proxy_stdout += rtn[0]
            if rtn[1]:
                proxy_stderr += rtn[1]
        return ProcProxy(proxy_stdout, proxy_stderr)
    else:
        # handles uncaptured mode
        rtn = alias(args, stdin=stdin)
        rtnout, rtnerr = None, None
        if isinstance(rtn, str):
            rtnout = rtn
            sys.stdout.write(rtn)
        elif isinstance(rtn, Sequence):
            if rtn[0]:
                rtnout = rtn[0]
                sys.stdout.write(rtn[0])
            if rtn[1]:
                rtnerr = rtn[1]
                sys.stderr.write(rtn[1])
        return ProcProxy(rtnout, rtnerr)


RE_SHEBANG = re.compile(r'#![ \t]*(.+?)$')


def _is_runnable_name(fname):
    return os.path.isfile(fname) and fname != os.path.basename(fname)


def _is_binary(fname, limit=80):
    with open(fname, 'rb') as f:
        for i in range(limit):
            char = f.read(1)
            if char == b'\0':
                return True
            if char == b'\n':
                return False
            if char == b'':
                return False
    return False


def get_script_subproc_command(fname, args):
    """
    Given the name of a script outside the path, returns a list representing
    an appropriate subprocess command to execute the script.  Raises
    PermissionError if the script is not executable.
    """
    # make sure file is executable
    if not os.access(fname, os.X_OK):
        raise PermissionError

    # if the file is a binary, we should call it directly
    if _is_binary(fname):
        return [fname] + args

    # find interpreter
    with open(fname, 'rb') as f:
        first_line = f.readline().decode().strip()
    m = RE_SHEBANG.match(first_line)

    # xonsh is the default interpreter
    if m is None:
        interp = ['xonsh']
    else:
        interp = m.group(1).strip()
        if len(interp) > 0:
            interp = shlex.split(interp)
        else:
            interp = ['xonsh']

    return interp + [fname] + args


def _subproc_pre():
    os.setpgrp()
    signal.signal(signal.SIGTSTP, lambda n, f: signal.pause())


def run_subproc(cmds, captured=True):
    """Runs a subprocess, in its many forms. This takes a list of 'commands,'
    which may be a list of command line arguments or a string, representing
    a special connecting character.  For example::

        $ ls | grep wakka

    is represented by the following cmds::

        [['ls'], '|', ['grep', 'wakka']]

    Lastly, the captured argument affects only the last real command.
    """
    global ENV
    last_stdout = PIPE if captured else None
    background = False
    if cmds[-1] == '&':
        background = True
        cmds = cmds[:-1]
    write_target = None
    if len(cmds) >= 3 and cmds[-2] in WRITER_MODES:
        write_target = cmds[-1][0]
        write_mode = WRITER_MODES[cmds[-2]]
        cmds = cmds[:-2]
        if write_target is not None:
            try:
                last_stdout = open(write_target, write_mode)
            except FileNotFoundError:
                e = 'xonsh: {0}: no such file or directory'
                raise XonshError(e.format(write_target))
        else:
            last_stdout = PIPE
    last_cmd = cmds[-1]
    prev = None
    procs = []
    prev_proc = None
    for cmd in cmds:
        if isinstance(cmd, string_types):
            prev = cmd
            continue
        stdout = last_stdout if cmd is last_cmd else PIPE
        uninew = cmd is last_cmd
        alias = builtins.aliases.get(cmd[0], None)
        if _is_runnable_name(cmd[0]):
            try:
                aliased_cmd = get_script_subproc_command(cmd[0], cmd[1:])
            except PermissionError:
                e = 'xonsh: subprocess mode: permission denied: {0}'
                raise XonshError(e.format(cmd[0]))
        elif alias is None:
            aliased_cmd = cmd
        elif callable(alias):
            prev_proc = _run_callable_subproc(alias, cmd[1:],
                                              captured=captured,
                                              prev_proc=prev_proc,
                                              stdout=stdout)
            continue
        else:
            aliased_cmd = alias + cmd[1:]
        # compute stdin for subprocess
        prev_is_proxy = isinstance(prev_proc, ProcProxy)
        if prev_proc is None:
            stdin = None
        elif prev_is_proxy:
            stdin = PIPE
        else:
            stdin = prev_proc.stdout
        subproc_kwargs = {}
        if os.name == 'posix':
            subproc_kwargs['preexec_fn'] = _subproc_pre
        try:
            proc = Popen(aliased_cmd,
                         universal_newlines=uninew,
                         env=ENV.detype(),
                         stdin=stdin,
                         stdout=stdout, **subproc_kwargs)
        except PermissionError:
            e = 'xonsh: subprocess mode: permission denied: {0}'
            raise XonshError(e.format(aliased_cmd[0]))
        except FileNotFoundError:
            cmd = aliased_cmd[0]
            e = 'xonsh: subprocess mode: command not found: {0}'.format(cmd)
            e += '\n' + suggest_commands(cmd, ENV, builtins.aliases)
            raise XonshError(e)
        procs.append(proc)
        prev = None
        if prev_is_proxy:
            proc.stdin.write(prev_proc.stdout)
            proc.stdin.close()
        prev_proc = proc
    for proc in procs[:-1]:
        proc.stdout.close()
    num = get_next_job_number()
    pids = [i.pid for i in procs]
    if not isinstance(prev_proc, ProcProxy):
        builtins.__xonsh_active_job__ = num
        builtins.__xonsh_all_jobs__[num] = {
            'cmds': cmds,
            'pids': pids,
            'obj': prev_proc,
            'started': time.time(),
            'pgrp': os.getpgid(prev_proc.pid),
            'status': 'running',
            'bg': background
        }
    if background:
        print_one_job(num)
        return
    wait_for_active_job()
    if write_target is None:
        # get output
        if isinstance(prev_proc, ProcProxy):
            output = prev_proc.stdout
        elif prev_proc.stdout is not None:
            output = prev_proc.stdout.read()
        if captured:
            return output
    elif last_stdout not in (PIPE, None):
        last_stdout.close()


def subproc_captured(*cmds):
    """Runs a subprocess, capturing the output. Returns the stdout
    that was produced as a str.
    """
    return run_subproc(cmds, captured=True)


def subproc_uncaptured(*cmds):
    """Runs a subprocess, without capturing the output. Returns the stdout
    that was produced as a str.
    """
    return run_subproc(cmds, captured=False)


def ensure_list_of_strs(x):
    """Ensures that x is a list of strings."""
    if isinstance(x, string_types):
        rtn = [x]
    elif isinstance(x, Sequence):
        rtn = [i if isinstance(i, string_types) else str(i) for i in x]
    else:
        rtn = [str(x)]
    return rtn


def load_builtins(execer=None):
    """Loads the xonsh builtins into the Python builtins. Sets the
    BUILTINS_LOADED variable to True.
    """
    global BUILTINS_LOADED, ENV
    # private built-ins
    builtins.__xonsh_env__ = ENV = Env(default_env())
    builtins.__xonsh_help__ = helper
    builtins.__xonsh_superhelp__ = superhelper
    builtins.__xonsh_regexpath__ = regexpath
    builtins.__xonsh_glob__ = globpath
    builtins.__xonsh_exit__ = False
    if hasattr(builtins, 'exit'):
        builtins.__xonsh_pyexit__ = builtins.exit
        del builtins.exit
    if hasattr(builtins, 'quit'):
        builtins.__xonsh_pyquit__ = builtins.quit
        del builtins.quit
    builtins.__xonsh_subproc_captured__ = subproc_captured
    builtins.__xonsh_subproc_uncaptured__ = subproc_uncaptured
    builtins.__xonsh_execer__ = execer
    builtins.__xonsh_all_jobs__ = {}
    builtins.__xonsh_active_job__ = None
    builtins.__xonsh_ensure_list_of_strs__ = ensure_list_of_strs
    # public built-ins
    builtins.evalx = None if execer is None else execer.eval
    builtins.execx = None if execer is None else execer.exec
    builtins.compilex = None if execer is None else execer.compile
    builtins.default_aliases = builtins.aliases = Aliases(DEFAULT_ALIASES)
    builtins.aliases.update(bash_aliases())
    BUILTINS_LOADED = True


def unload_builtins():
    """Removes the xonsh builtins from the Python builtins, if the
    BUILTINS_LOADED is True, sets BUILTINS_LOADED to False, and returns.
    """
    global BUILTINS_LOADED, ENV
    if ENV is not None:
        ENV.undo_replace_env()
        ENV = None
    if hasattr(builtins, '__xonsh_pyexit__'):
        builtins.exit = builtins.__xonsh_pyexit__
    if hasattr(builtins, '__xonsh_pyquit__'):
        builtins.quit = builtins.__xonsh_pyquit__
    if not BUILTINS_LOADED:
        return
    names = ['__xonsh_env__',
             '__xonsh_help__',
             '__xonsh_superhelp__',
             '__xonsh_regexpath__',
             '__xonsh_glob__',
             '__xonsh_exit__',
             '__xonsh_pyexit__',
             '__xonsh_pyquit__',
             '__xonsh_subproc_captured__',
             '__xonsh_subproc_uncaptured__',
             '__xonsh_execer__',
             'evalx',
             'execx',
             'compilex',
             'default_aliases',
             '__xonsh_all_jobs__',
             '__xonsh_active_job__',
             '__xonsh_ensure_list_of_strs__', ]
    for name in names:
        if hasattr(builtins, name):
            delattr(builtins, name)
    BUILTINS_LOADED = False


@contextmanager
def xonsh_builtins(execer=None):
    """A context manager for using the xonsh builtins only in a limited
    scope. Likely useful in testing.
    """
    load_builtins(execer=execer)
    yield
    unload_builtins()



"""Directory stack and associated utilities for the xonsh shell.
"""
import os
import builtins
from argparse import ArgumentParser

DIRSTACK = []
"""
A list containing the currently remembered directories.
"""


def pushd(args, stdin=None):
    """
    xonsh command: pushd

    Adds a directory to the top of the directory stack, or rotates the stack,
    making the new top of the stack the current working directory.
    """
    global DIRSTACK

    try:
        args = pushd_parser.parse_args(args)
    except SystemExit:
        return None, None

    env = builtins.__xonsh_env__

    pwd = env['PWD']

    if env.get('PUSHD_MINUS', False):
        BACKWARD = '-'
        FORWARD = '+'
    else:
        BACKWARD = '+'
        FORWARD = '-'

    if args.dir is None:
        try:
            new_pwd = DIRSTACK.pop(0)
        except IndexError:
            e = 'pushd: Directory stack is empty\n'
            return None, e
    elif os.path.isdir(args.dir):
        new_pwd = args.dir
    else:
        try:
            num = int(args.dir[1:])
        except ValueError:
            e = 'Invalid argument to pushd: {0}\n'
            return None, e.format(args.dir)

        if num < 0:
            e = 'Invalid argument to pushd: {0}\n'
            return None, e.format(args.dir)

        if num > len(DIRSTACK):
            e = 'Too few elements in dirstack ({0} elements)\n'
            return None, e.format(len(DIRSTACK))
        elif args.dir.startswith(FORWARD):
            if num == len(DIRSTACK):
                new_pwd = None
            else:
                new_pwd = DIRSTACK.pop(len(DIRSTACK) - 1 - num)
        elif args.dir.startswith(BACKWARD):
            if num == 0:
                new_pwd = None
            else:
                new_pwd = DIRSTACK.pop(num - 1)
        else:
            e = 'Invalid argument to pushd: {0}\n'
            return None, e.format(args.dir)
    if new_pwd is not None:
        e = None
        if args.cd:
            DIRSTACK.insert(0, os.path.expanduser(pwd))
            _, e = builtins.default_aliases['cd']([new_pwd], None)
        else:
            DIRSTACK.insert(0, os.path.expanduser(os.path.abspath(new_pwd)))

        if e is not None:
            return None, e

    maxsize = env.get('DIRSTACK_SIZE', 20)
    if len(DIRSTACK) > maxsize:
        DIRSTACK = DIRSTACK[:maxsize]

    if not args.quiet and not env.get('PUSHD_SILENT', False):
        return dirs([], None)

    return None, None


def popd(args, stdin=None):
    """
    xonsh command: popd

    Removes entries from the directory stack.
    """
    global DIRSTACK

    try:
        args = pushd_parser.parse_args(args)
    except SystemExit:
        return None, None

    env = builtins.__xonsh_env__

    if env.get('PUSHD_MINUS', False):
        BACKWARD = '-'
        FORWARD = '+'
    else:
        BACKWARD = '-'
        FORWARD = '+'

    if args.dir is None:
        try:
            new_pwd = DIRSTACK.pop(0)
        except IndexError:
            e = 'popd: Directory stack is empty\n'
            return None, e
    else:
        try:
            num = int(args.dir[1:])
        except ValueError:
            e = 'Invalid argument to popd: {0}\n'
            return None, e.format(args.dir)

        if num < 0:
            e = 'Invalid argument to popd: {0}\n'
            return None, e.format(args.dir)

        if num > len(DIRSTACK):
            e = 'Too few elements in dirstack ({0} elements)\n'
            return None, e.format(len(DIRSTACK))
        elif args.dir.startswith(FORWARD):
            if num == len(DIRSTACK):
                new_pwd = DIRSTACK.pop(0)
            else:
                new_pwd = None
                DIRSTACK.pop(len(DIRSTACK) - 1 - num)
        elif args.dir.startswith(BACKWARD):
            if num == 0:
                new_pwd = DIRSTACK.pop(0)
            else:
                new_pwd = None
                DIRSTACK.pop(num - 1)
        else:
            e = 'Invalid argument to popd: {0}\n'
            return None, e.format(args.dir)

    if new_pwd is not None:
        e = None
        if args.cd:
            _, e = builtins.default_aliases['cd']([new_pwd], None)

        if e is not None:
            return None, e

    if not args.quiet and not env.get('PUSHD_SILENT', False):
        return dirs([], None)

    return None, None


def dirs(args, stdin=None):
    """
    xonsh command: dirs

    Displays the list of currently remembered directories.  Can also be used
    to clear the directory stack.
    """
    global DIRSTACK
    dirstack = [os.path.expanduser(builtins.__xonsh_env__['PWD'])] + DIRSTACK

    try:
        args = dirs_parser.parse_args(args)
    except SystemExit:
        return None, None

    env = builtins.__xonsh_env__

    if env.get('PUSHD_MINUS', False):
        BACKWARD = '-'
        FORWARD = '+'
    else:
        BACKWARD = '-'
        FORWARD = '+'

    if args.clear:
        dirstack = []
        return None, None

    if args.long:
        o = dirstack
    else:
        d = os.path.expanduser('~')
        o = [i.replace(d, '~') for i in dirstack]

    if args.verbose:
        out = ''
        pad = len(str(len(o) - 1))
        for (ix, e) in enumerate(o):
            blanks = ' ' * (pad - len(str(ix)))
            out += '\n{0}{1} {2}'.format(blanks, ix, e)
        out = out[1:]
    elif args.print_long:
        out = '\n'.join(o)
    else:
        out = ' '.join(o)

    N = args.N
    if N is not None:
        try:
            num = int(N[1:])
        except ValueError:
            e = 'Invalid argument to dirs: {0}\n'
            return None, e.format(N)

        if num < 0:
            e = 'Invalid argument to dirs: {0}\n'
            return None, e.format(len(o))

        if num >= len(o):
            e = 'Too few elements in dirstack ({0} elements)\n'
            return None, e.format(len(o))

        if N.startswith(BACKWARD):
            idx = num
        elif N.startswith(FORWARD):
            idx = len(o) - 1 - num
        else:
            e = 'Invalid argument to dirs: {0}\n'
            return None, e.format(N)

        out = o[idx]

    return out + '\n', None


pushd_parser = ArgumentParser(prog="pushd")
pushd_parser.add_argument('dir', nargs='?')
pushd_parser.add_argument('-n',
                          dest='cd',
                          help='Suppresses the normal change of directory when'
                          ' adding directories to the stack, so that only the'
                          ' stack is manipulated.',
                          action='store_false')
pushd_parser.add_argument('-q',
                          dest='quiet',
                          help='Do not call dirs, regardless of $PUSHD_SILENT',
                          action='store_true')

popd_parser = ArgumentParser(prog="popd")
popd_parser.add_argument('dir', nargs='?')
popd_parser.add_argument('-n',
                         dest='cd',
                         help='Suppresses the normal change of directory when'
                         ' adding directories to the stack, so that only the'
                         ' stack is manipulated.',
                         action='store_false')
popd_parser.add_argument('-q',
                         dest='quiet',
                         help='Do not call dirs, regardless of $PUSHD_SILENT',
                         action='store_true')

dirs_parser = ArgumentParser(prog="dirs")
dirs_parser.add_argument('N', nargs='?')
dirs_parser.add_argument('-c',
                         dest='clear',
                         help='Clears the directory stack by deleting all of'
                         ' the entries.',
                         action='store_true')
dirs_parser.add_argument('-p',
                         dest='print_long',
                         help='Print the directory stack with one entry per'
                         ' line.',
                         action='store_true')
dirs_parser.add_argument('-v',
                         dest='verbose',
                         help='Print the directory stack with one entry per'
                         ' line, prefixing each entry with its index in the'
                         ' stack.',
                         action='store_true')
dirs_parser.add_argument('-l',
                         dest='long',
                         help='Produces a longer listing; the default listing'
                         ' format uses a tilde to denote the home directory.',
                         action='store_true')



"""Aliases for the xonsh shell.
"""
import os
import platform
import builtins
import subprocess
import shlex
from warnings import warn

from xonsh.dirstack import dirs, pushd, popd
from xonsh.jobs import jobs, fg, bg, kill_all_jobs


def cd(args, stdin=None):
    """Changes the directory.

    If no directory is specified (i.e. if `args` is None) then this
    changes to the current user's home directory.
    """
    env = builtins.__xonsh_env__
    cur_oldpwd = env.get('OLDPWD', os.getcwd())
    if len(args) == 0:
        d = os.path.expanduser('~')
    elif len(args) == 1:
        d = os.path.expanduser(args[0])
        if d == '-':
            d = cur_oldpwd
    else:
        return '', 'cd takes 0 or 1 arguments, not {0}\n'.format(len(args))
    if not os.path.exists(d):
        return '', 'cd: no such file or directory: {0}\n'.format(d)
    if not os.path.isdir(d):
        return '', 'cd: {0} is not a directory\n'.format(d)

    env['OLDPWD'] = os.getcwd()
    os.chdir(d)
    env['PWD'] = os.getcwd()
    return None, None


def exit(args, stdin=None):  # pylint:disable=redefined-builtin,W0622
    """Sends signal to exit shell."""
    builtins.__xonsh_exit__ = True
    kill_all_jobs()
    print()  # gimme a newline
    return None, None


def source_bash(args, stdin=None):
    """Implements bash's source builtin."""
    import tempfile
    env = builtins.__xonsh_env__
    denv = env.detype()
    with tempfile.NamedTemporaryFile(mode='w+t') as f:
        args = ' '.join(args)
        inp = 'source {0}\nenv >> {1}\n'.format(args, f.name)
        try:
            subprocess.check_output(['bash'],
                                    input=inp,
                                    env=denv,
                                    stderr=subprocess.PIPE,
                                    universal_newlines=True)
        except subprocess.CalledProcessError:
            return None, 'could not source {0}\n'.format(args)
        f.seek(0)
        exported = f.read()
    items = [l.split('=', 1) for l in exported.splitlines() if '=' in l]
    newenv = dict(items)
    for k, v in newenv.items():
        if k in env and v == denv[k]:
            continue  # no change from original
        env[k] = v
    return


def bash_aliases():
    """Computes a dictionary of aliases based on Bash's aliases."""
    try:
        s = subprocess.check_output(['bash', '-i'],
                                    input='alias',
                                    stderr=subprocess.PIPE,
                                    universal_newlines=True)
    except subprocess.CalledProcessError:
        s = ''
    items = [line.split('=', 1) for line in s.splitlines() if '=' in line]
    aliases = {}
    for key, value in items:
        try:
            key = key[6:]
            value = value.strip('\'')
            value = shlex.split(value)
        except ValueError as exc:
            warn('could not parse Bash alias "{0}": {1!r}'.format(key, exc),
                 RuntimeWarning)
            continue
        aliases[key] = value
    return aliases


DEFAULT_ALIASES = {
    'cd': cd,
    'pushd': pushd,
    'popd': popd,
    'dirs': dirs,
    'jobs': jobs,
    'fg': fg,
    'bg': bg,
    'EOF': exit,
    'exit': exit,
    'quit': exit,
    'source-bash': source_bash,
    'grep': ['grep', '--color=auto'],
    'scp-resume': ['rsync', '--partial', '-h', '--progress', '--rsh=ssh'],
    'ipynb': ['ipython', 'notebook', '--no-browser'],
}

if platform.system() == 'Darwin':
    DEFAULT_ALIASES['ls'] = ['ls', '-G']
else:
    DEFAULT_ALIASES['ls'] = ['ls', '--color=auto', '-v']



"""Implements the xonsh parser"""
from __future__ import print_function, unicode_literals
import os
from collections import Iterable, Sequence, Mapping

from ply import yacc

from xonsh import ast
from xonsh.lexer import Lexer


class Location(object):
    """Location in a file."""

    def __init__(self, fname, lineno, column=None):
        """Takes a filename, line number, and optionally a column number."""
        self.fname = fname
        self.lineno = lineno
        self.column = column

    def __str__(self):
        s = '{0}:{1}'.format(self.fname, self.lineno)
        if self.column is not None:
            s += ':{0}'.format(self.column)
        return s


def has_elts(x):
    """Tests if x is an AST node with elements."""
    return isinstance(x, ast.AST) and hasattr(x, 'elts')


def ensure_has_elts(x, lineno=1, col_offset=1):
    """Ensures that x is an AST node with elements."""
    if not has_elts(x):
        if not isinstance(x, Iterable):
            x = [x]
        x = ast.Tuple(elts=x,
                      ctx=ast.Load(),
                      lineno=lineno,
                      col_offset=col_offset)
    return x


def empty_list(lineno=None, col=None):
    """Creates the AST node for an empty list."""
    return ast.List(elts=[], ctx=ast.Load(), lineno=lineno, col_offset=col)


def binop(x, op, y, lineno=None, col=None):
    """Creates the AST node for a binary operation."""
    return ast.BinOp(left=x, op=op, right=y, lineno=lineno, col_offset=col)


def call_split_lines(x, lineno=None, col=None):
    """Creates the AST node for calling the 'splitlines' attribute of an
    object, nominally a string.
    """
    return ast.Call(func=ast.Attribute(value=x,
                                       attr='splitlines',
                                       ctx=ast.Load(),
                                       lineno=lineno,
                                       col_offset=col),
                    args=[],
                    keywords=[],
                    starargs=None,
                    kwargs=None,
                    lineno=lineno,
                    col_offset=col)


def ensure_list_from_str_or_list(x, lineno=None, col=None):
    """Creates the AST node for the following expression::

        [x] if isinstance(x, str) else x

    Somewhat useful.
    """
    return ast.IfExp(test=ast.Call(func=ast.Name(id='isinstance',
                                                 ctx=ast.Load(),
                                                 lineno=lineno,
                                                 col_offset=col),
                                   args=[x, ast.Name(id='str',
                                                     ctx=ast.Load(),
                                                     lineno=lineno,
                                                     col_offset=col)],
                                   keywords=[],
                                   starargs=None,
                                   kwargs=None,
                                   lineno=lineno,
                                   col_offset=col),
                     body=ast.List(elts=[x],
                                   ctx=ast.Load(),
                                   lineno=lineno,
                                   col_offset=col),
                     orelse=x,
                     lineno=lineno,
                     col_offset=col)


def xonsh_call(name, args, lineno=None, col=None):
    """Creates the AST node for calling a function of a given name."""
    return ast.Call(func=ast.Name(id=name,
                                  ctx=ast.Load(),
                                  lineno=lineno,
                                  col_offset=col),
                    args=args,
                    keywords=[],
                    starargs=None,
                    kwargs=None,
                    lineno=lineno,
                    col_offset=col)


def xonsh_help(x, lineno=None, col=None):
    """Creates the AST node for calling the __xonsh_help__() function."""
    return xonsh_call('__xonsh_help__', [x], lineno=lineno, col=col)


def xonsh_superhelp(x, lineno=None, col=None):
    """Creates the AST node for calling the __xonsh_superhelp__() function."""
    return xonsh_call('__xonsh_superhelp__', [x], lineno=lineno, col=col)


def xonsh_regexpath(x, lineno=None, col=None):
    """Creates the AST node for calling the __xonsh_regexpath__() function."""
    return xonsh_call('__xonsh_regexpath__', [x], lineno=lineno, col=col)


def load_ctx(x):
    """Recursively sets ctx to ast.Load()"""
    if not hasattr(x, 'ctx'):
        return
    x.ctx = ast.Load()
    if isinstance(x, (ast.Tuple, ast.List)):
        for e in x.elts:
            load_ctx(e)
    elif isinstance(x, ast.Starred):
        load_ctx(x.value)


def store_ctx(x):
    """Recursively sets ctx to ast.Store()"""
    if not hasattr(x, 'ctx'):
        return
    x.ctx = ast.Store()
    if isinstance(x, (ast.Tuple, ast.List)):
        for e in x.elts:
            store_ctx(e)
    elif isinstance(x, ast.Starred):
        store_ctx(x.value)


def empty_list_if_newline(x):
    return [] if x == '\n' else x


class Parser(object):
    """A class that parses the xonsh language."""

    def __init__(self,
                 lexer_optimize=True,
                 lexer_table='xonsh.lexer_table',
                 yacc_optimize=True,
                 yacc_table='xonsh.parser_table',
                 yacc_debug=False,
                 outputdir=None):
        """Parameters
        ----------
        lexer_optimize : bool, optional
            Set to false when unstable and true when lexer is stable.
        lexer_table : str, optional
            Lexer module used when optimized.
        yacc_optimize : bool, optional
            Set to false when unstable and true when parser is stable.
        yacc_table : str, optional
            Parser module used when optimized.
        yacc_debug : debug, optional
            Dumps extra debug info.
        outputdir : str or None, optional
            The directory to place generated tables within.
        """
        self.lexer = lexer = Lexer()
        self.tokens = lexer.tokens

        opt_rules = (
            'newlines', 'arglist', 'func_call', 'rarrow_test', 'typedargslist',
            'equals_test', 'colon_test', 'tfpdef', 'comma_tfpdef_list',
            'comma_pow_tfpdef', 'vfpdef', 'comma_vfpdef_list',
            'comma_pow_vfpdef', 'equals_yield_expr_or_testlist_list',
            'testlist', 'as_name', 'period_or_ellipsis_list',
            'comma_import_as_name_list', 'comma_dotted_as_name_list',
            'comma_name_list', 'comma_test', 'elif_part_list', 'finally_part',
            'varargslist', 'or_and_test_list', 'and_not_test_list',
            'comp_op_expr_list', 'xor_and_expr_list',
            'ampersand_shift_expr_list', 'shift_arith_expr_list',
            'op_factor_list', 'trailer_list', 'testlist_comp',
            'yield_expr_or_testlist_comp', 'dictorsetmaker',
            'comma_subscript_list', 'test', 'sliceop', 'comp_iter',
            'yield_arg', 'argument_comma_list', 'comma_argument_list',
            'test_comma_list', )
        for rule in opt_rules:
            self._opt_rule(rule)

        list_rules = (
            'comma_tfpdef', 'comma_vfpdef', 'semi_small_stmt',
            'comma_test_or_star_expr', 'period_or_ellipsis',
            'comma_import_as_name', 'comma_dotted_as_name', 'period_name',
            'comma_name', 'elif_part', 'except_part', 'comma_with_item',
            'or_and_test', 'and_not_test', 'comp_op_expr', 'pipe_xor_expr',
            'xor_and_expr', 'ampersand_shift_expr', 'shift_arith_expr',
            'pm_term', 'op_factor', 'trailer', 'comma_subscript',
            'comma_expr_or_star_expr', 'comma_test', 'argument_comma',
            'comma_argument', 'comma_item', 'attr_period_name', 'test_comma',
            'equals_yield_expr_or_testlist', )
        for rule in list_rules:
            self._list_rule(rule)

        yacc_kwargs = dict(module=self,
                           debug=yacc_debug,
                           start='start_symbols',
                           optimize=yacc_optimize,
                           tabmodule=yacc_table)
        if not yacc_debug:
            yacc_kwargs['errorlog'] = yacc.NullLogger()
        if outputdir is not None:
            yacc_kwargs['outputdir'] = outputdir
        self.parser = yacc.yacc(**yacc_kwargs)

        # Keeps track of the last token given to yacc (the lookahead token)
        self._last_yielded_token = None

    def reset(self):
        """Resets for clean parsing."""
        self.lexer.reset()
        self._last_yielded_token = None

    def parse(self, s, filename='<code>', mode='exec', debug_level=0):
        """Returns an abstract syntax tree of xonsh code.

        Parameters
        ----------
        s : str
            The xonsh code.
        filename : str, optional
            Name of the file.
        mode : str, optional
            Execution mode, one of: exec, eval, or single.
        debug_level : str, optional
            Debugging level passed down to yacc.

        Returns
        -------
        tree : AST
        """
        self.reset()
        self.lexer.fname = filename
        tree = self.parser.parse(input=s, lexer=self.lexer, debug=debug_level)
        # hack for getting modes right
        if mode == 'single':
            if isinstance(tree, ast.Expression):
                tree = ast.Interactive(body=[self.expr(tree.body)])
            elif isinstance(tree, ast.Module):
                tree = ast.Interactive(body=tree.body)
        return tree

    def _lexer_errfunc(self, msg, line, column):
        self._parse_error(msg, self.currloc(line, column))

    def _yacc_lookahead_token(self):
        """Gets the last token seen by the lexer."""
        return self.lexer.last

    def _opt_rule(self, rulename):
        """For a rule name, creates an associated optional rule.
        '_opt' is appended to the rule name.
        """

        def optfunc(self, p):
            p[0] = p[1]

        optfunc.__doc__ = ('{0}_opt : empty\n'
                           '        | {0}').format(rulename)
        optfunc.__name__ = 'p_' + rulename + '_opt'
        setattr(self.__class__, optfunc.__name__, optfunc)

    def _list_rule(self, rulename):
        """For a rule name, creates an associated list rule.
        '_list' is appended to the rule name.
        """

        def listfunc(self, p):
            p[0] = p[1] if len(p) == 2 else p[1] + p[2]

        listfunc.__doc__ = ('{0}_list : {0}\n'
                            '         | {0}_list {0}').format(rulename)
        listfunc.__name__ = 'p_' + rulename + '_list'
        setattr(self.__class__, listfunc.__name__, listfunc)

    def currloc(self, lineno, column=None):
        """Returns the current location."""
        return Location(fname=self.lexer.fname, lineno=lineno, column=column)

    def expr(self, p):
        """Creates an expression for a token."""
        return ast.Expr(value=p, lineno=p.lineno, col_offset=p.col_offset)

    def token_col(self, t):
        """Gets ths token column"""
        return t.lexpos

    @property
    def lineno(self):
        if self.lexer.last is None:
            return 0
        else:
            return self.lexer.last.lineno

    @property
    def col(self):
        t = self._yacc_lookahead_token()
        if t is not None:
            return self.token_col(t)
        return 1

    def _parse_error(self, msg, loc):
        err = SyntaxError('{0}: {1}'.format(loc, msg))
        err.loc = loc
        raise err

    #
    # Precedence of operators
    #
    precedence = (('left', 'PIPE'), ('left', 'XOR'), ('left', 'AMPERSAND'),
                  ('left', 'EQ', 'NE'), ('left', 'GT', 'GE', 'LT', 'LE'),
                  ('left', 'RSHIFT', 'LSHIFT'), ('left', 'PLUS', 'MINUS'),
                  ('left', 'TIMES', 'DIVIDE', 'DOUBLEDIV', 'MOD'),
                  ('left', 'POW'), )

    #
    # Grammar as defined by BNF
    #

    def p_start_symbols(self, p):
        """start_symbols : single_input
                         | file_input
                         | eval_input
                         | empty
        """
        p[0] = p[1]

    def p_single_input(self, p):
        """single_input : compound_stmt NEWLINE
        """
        p1 = empty_list_if_newline(p[1])
        p0 = ast.Interactive(body=p1)
        p[0] = p0

    def p_file_input(self, p):
        """file_input : file_stmts"""
        p[0] = ast.Module(body=p[1])

    def p_file_stmts(self, p):
        """file_stmts : newline_or_stmt
                      | file_stmts newline_or_stmt
        """
        if len(p) == 2:
            # newline_or_stmt ENDMARKER
            p1 = empty_list_if_newline(p[1])
            p[0] = p1
        else:
            # file_input newline_or_stmt ENDMARKER
            p2 = empty_list_if_newline(p[2])
            p[0] = p[1] + p2

    def p_newline_or_stmt(self, p):
        """newline_or_stmt : NEWLINE
                           | stmt
        """
        p[0] = p[1]

    def p_newlines(self, p):
        """newlines : NEWLINE
                    | newlines NEWLINE
        """
        p[0] = p[1] if len(p) == 2 else p[1] + p[2]

    def p_eval_input(self, p):
        """eval_input : testlist newlines_opt
        """
        p[0] = ast.Expression(body=p[1])

    def p_func_call(self, p):
        """func_call : LPAREN arglist_opt RPAREN"""
        p[0] = p[2]

    def p_attr_period_name(self, p):
        """attr_period_name : PERIOD NAME"""
        p[0] = [p[2]]

    def p_attr_name(self, p):
        """attr_name : NAME
                     | NAME attr_period_name_list
        """
        p1 = p[1]
        name = ast.Name(id=p1,
                        ctx=ast.Load(),
                        lineno=self.lineno,
                        col_offset=self.col)
        if len(p) == 2:
            p0 = name
        else:
            p2 = p[2]
            p0 = ast.Attribute(value=name,
                               attr=p2[0],
                               ctx=ast.Load(),
                               lineno=self.lineno,
                               col_offset=self.col)
            for a in p2[1:]:
                p0 = ast.Attribute(value=p0,
                                   attr=a,
                                   ctx=ast.Load(),
                                   lineno=self.lineno,
                                   col_offset=self.col)
        p[0] = p0

    def p_decorator(self, p):
        """decorator : AT attr_name NEWLINE
                     | AT attr_name func_call NEWLINE
        """
        lenp = len(p)
        name = p[2]
        p3 = p[3] if lenp > 3 else None
        if lenp == 4:
            p0 = name
        elif p3 is None:
            p0 = ast.Call(func=name,
                          args=[],
                          keywords=[],
                          starargs=None,
                          kwargs=None,
                          lineno=self.lineno,
                          col_offset=self.col)
        else:
            p0 = ast.Call(func=name,
                          lineno=self.lineno,
                          col_offset=self.col, **p3)
        p[0] = p0

    def p_decorators(self, p):
        """decorators : decorator
                      | decorators decorator
        """
        p[0] = [p[1]] if len(p) == 2 else p[1] + [p[2]]

    def p_classdef_or_funcdef(self, p):
        """classdef_or_funcdef : classdef
                               | funcdef
        """
        p[0] = p[1]

    def p_decorated(self, p):
        """decorated : decorators classdef_or_funcdef"""
        p1, p2 = p[1], p[2]
        p2[0].decorator_list = p1
        p[0] = p2

    def p_rarrow_test(self, p):
        """rarrow_test : RARROW test"""
        p[0] = p[2]

    def p_funcdef(self, p):
        """funcdef : DEF NAME parameters rarrow_test_opt COLON suite"""
        f = ast.FunctionDef(name=p[2],
                            args=p[3],
                            returns=p[4],
                            body=p[6],
                            decorator_list=[],
                            lineno=self.lineno,
                            col_offset=self.col)
        p[0] = [f]

    def p_parameters(self, p):
        """parameters : LPAREN typedargslist_opt RPAREN"""
        p2 = p[2]
        if p2 is None:
            p2 = ast.arguments(args=[],
                               vararg=None,
                               kwonlyargs=[],
                               kw_defaults=[],
                               kwarg=None,
                               defaults=[])
        p[0] = p2

    def p_equals_test(self, p):
        """equals_test : EQUALS test"""
        p[0] = p[2]

    def p_typedargslist(self, p):
        """typedargslist : tfpdef equals_test_opt comma_tfpdef_list_opt comma_opt
                         | tfpdef equals_test_opt comma_tfpdef_list_opt comma_opt TIMES tfpdef_opt COMMA POW vfpdef
                         | tfpdef equals_test_opt comma_tfpdef_list_opt comma_opt TIMES tfpdef_opt comma_tfpdef_list_opt
                         | tfpdef equals_test_opt comma_tfpdef_list_opt comma_opt TIMES tfpdef_opt comma_tfpdef_list COMMA POW tfpdef
                         | tfpdef equals_test_opt comma_tfpdef_list_opt comma_opt POW tfpdef
                         | TIMES tfpdef_opt comma_tfpdef_list comma_pow_tfpdef_opt
                         | TIMES tfpdef_opt comma_pow_tfpdef_opt
                         | POW tfpdef
        """
        lenp = len(p)
        p1, p2 = p[1], p[2]
        p3 = p[3] if lenp > 3 else None
        p4 = p[4] if lenp > 4 else None
        # skip p5 
        p6 = p[6] if lenp > 6 else None
        p7 = p[7] if lenp > 7 else None
        # skip p8
        p9 = p[9] if lenp > 9 else None
        p10 = p[10] if lenp > 10 else None
        p0 = ast.arguments(args=[],
                           vararg=None,
                           kwonlyargs=[],
                           kw_defaults=[],
                           kwarg=None,
                           defaults=[])
        if lenp == 3:
            p0.kwarg = p2
        elif lenp == 4:
            self._set_var_args(p0, p2, None)
            p0.kwarg = p3
        elif lenp == 5 and p1 != '*':
            # x
            self._set_regular_args(p0, p1, p2, p3, p4)
        elif lenp == 5 and p1 == '*':
            self._set_var_args(p0, p2, p3)  # *args
            if p4 is not None:
                # *args, x, **kwargs
                p0.kwarg = p4
        elif lenp == 7:
            # x, **kwargs
            self._set_regular_args(p0, p1, p2, p3, p4)
            p0.kwarg = p[6]
        elif lenp == 8:
            # x, *args
            self._set_regular_args(p0, p1, p2, p3, p4)
            self._set_var_args(p0, p6, p7)
        elif lenp == 10:
            # x, *args, **kwargs
            self._set_regular_args(p0, p1, p2, p3, p4)
            self._set_var_args(p0, p6, None)
            p0.kwarg = p9
        elif lenp == 11:
            # x, *args, **kwargs
            self._set_regular_args(p0, p1, p2, p3, p4)
            self._set_var_args(p0, p6, p7)
            p0.kwarg = p10
        else:
            assert False
        p[0] = p0

    def p_colon_test(self, p):
        """colon_test : COLON test"""
        p[0] = p[2]

    def p_tfpdef(self, p):
        """tfpdef : NAME colon_test_opt"""
        p[0] = ast.arg(arg=p[1], annotation=p[2])

    def p_comma_tfpdef(self, p):
        """comma_tfpdef : COMMA
                        | COMMA tfpdef equals_test_opt
        """
        if len(p) == 2:
            p[0] = []
        else:
            p[0] = [{'arg': p[2], 'default': p[3]}]

    def p_comma_pow_tfpdef(self, p):
        """comma_pow_tfpdef : COMMA POW tfpdef"""
        p[0] = p[3]

    def _set_args_def(self, argmts, vals, kwargs=False):
        args, defs = (argmts.kwonlyargs, argmts.kw_defaults) if kwargs else \
                     (argmts.args, argmts.defaults)
        for v in vals:
            args.append(v['arg'])
            d = v['default']
            if kwargs or (d is not None):
                defs.append(d)

    def _set_regular_args(self, p0, p1, p2, p3, p4):
        if p2 is None and p3 is None:
            # x
            p0.args.append(p1)
        elif p2 is not None and p3 is None:
            # x=42
            p0.args.append(p1)
            p0.defaults.append(p2)
        elif p2 is None and p3 is not None:
            # x, y and x, y=42
            p0.args.append(p1)
            self._set_args_def(p0, p3)
        else:
            # x=42, y=42
            p0.args.append(p1)
            p0.defaults.append(p2)
            self._set_args_def(p0, p3)

    def _set_var_args(self, p0, vararg, kwargs):
        if vararg is None:
            self._set_args_def(p0, kwargs, kwargs=True)
        elif vararg is not None and kwargs is None:
            # *args
            p0.vararg = vararg
        elif vararg is not None and kwargs is not None:
            # *args, x and *args, x, y and *args, x=10 and *args, x=10, y
            # and *args, x, y=10, and *args, x=42, y=65
            p0.vararg = vararg
            self._set_args_def(p0, kwargs, kwargs=True)
        else:
            assert False

    def p_varargslist(self, p):
        """varargslist : vfpdef equals_test_opt comma_vfpdef_list_opt comma_opt
                       | vfpdef equals_test_opt comma_vfpdef_list_opt comma_opt TIMES vfpdef_opt COMMA POW vfpdef
                       | vfpdef equals_test_opt comma_vfpdef_list_opt comma_opt TIMES vfpdef_opt comma_vfpdef_list_opt
                       | vfpdef equals_test_opt comma_vfpdef_list_opt comma_opt TIMES vfpdef_opt comma_vfpdef_list COMMA POW vfpdef
                       | vfpdef equals_test_opt comma_vfpdef_list_opt comma_opt POW vfpdef
                       | TIMES vfpdef_opt comma_vfpdef_list comma_pow_vfpdef_opt
                       | TIMES vfpdef_opt comma_pow_vfpdef_opt
                       | POW vfpdef
        """
        lenp = len(p)
        p1, p2 = p[1], p[2]
        p3 = p[3] if lenp > 3 else None
        p4 = p[4] if lenp > 4 else None
        # skip p5
        p6 = p[6] if lenp > 6 else None
        p7 = p[7] if lenp > 7 else None
        # skip p8
        p9 = p[9] if lenp > 9 else None
        p10 = p[10] if lenp > 10 else None
        p0 = ast.arguments(args=[],
                           vararg=None,
                           kwonlyargs=[],
                           kw_defaults=[],
                           kwarg=None,
                           defaults=[])
        if lenp == 3:
            p0.kwarg = p2
        elif lenp == 4:
            self._set_var_args(p0, p2, None)
            p0.kwarg = p3
        elif lenp == 5 and p1 != '*':
            # x
            self._set_regular_args(p0, p1, p2, p3, p4)
        elif lenp == 5 and p1 == '*':
            self._set_var_args(p0, p2, p3)  # *args
            if p4 is not None:
                # *args, x, **kwargs
                p0.kwarg = p4
        elif lenp == 7:
            # x, **kwargs
            self._set_regular_args(p0, p1, p2, p3, p4)
            p0.kwarg = p6
        elif lenp == 8:
            # x, *args
            self._set_regular_args(p0, p1, p2, p3, p4)
            self._set_var_args(p0, p6, p7)
        elif lenp == 10:
            # x, *args, **kwargs
            self._set_regular_args(p0, p1, p2, p3, p4)
            self._set_var_args(p0, p6, None)
            p0.kwarg = p9
        elif lenp == 11:
            # x, *args, **kwargs
            self._set_regular_args(p0, p1, p2, p3, p4)
            self._set_var_args(p0, p6, p7)
            p0.kwarg = p10
        else:
            assert False
        p[0] = p0

    def p_vfpdef(self, p):
        """vfpdef : NAME"""
        p[0] = ast.arg(arg=p[1], annotation=None)

    def p_comma_vfpdef(self, p):
        """comma_vfpdef : COMMA
                        | COMMA vfpdef equals_test_opt
        """
        if len(p) == 2:
            p[0] = []
        else:
            p[0] = [{'arg': p[2], 'default': p[3]}]

    def p_comma_pow_vfpdef(self, p):
        """comma_pow_vfpdef : COMMA POW vfpdef"""
        p[0] = p[3]

    def p_stmt(self, p):
        """stmt : simple_stmt
                | compound_stmt
        """
        p[0] = p[1]

    def p_stmt_list(self, p):
        """stmt_list : stmt
                     | stmt_list stmt"""
        if len(p) == 2:
            p[0] = p[1]
        else:
            p[0] = p[1] + p[2]

    def p_semi_opt(self, p):
        """semi_opt : SEMI
                    | empty
        """
        if len(p) == 2:
            p[0] = p[1]

    def p_semi_small_stmt(self, p):
        """semi_small_stmt : SEMI small_stmt"""
        p[0] = [p[2]]

    def p_simple_stmt(self, p):
        """simple_stmt : small_stmt semi_small_stmt_list semi_opt NEWLINE
                       | small_stmt semi_opt NEWLINE
        """
        p1, p2 = p[1], p[2]
        p0 = [p1]
        if p2 is not None and p2 != ';':
            p0 += p2
        p[0] = p0

    def p_small_stmt(self, p):
        """small_stmt : expr_stmt
                      | del_stmt
                      | pass_stmt
                      | flow_stmt
                      | import_stmt
                      | global_stmt
                      | nonlocal_stmt
                      | assert_stmt
        """
        p[0] = p[1]

    _augassign_op = {
        '+=': ast.Add,
        '-=': ast.Sub,
        '*=': ast.Mult,
        '/=': ast.Div,
        '%=': ast.Mod,
        '//=': ast.FloorDiv,
        '**=': ast.Pow,
        '^=': ast.BitXor,
        '&=': ast.BitAnd,
        '|=': ast.BitOr,
        '<<=': ast.LShift,
        '>>=': ast.RShift
    }

    def p_expr_stmt(self, p):
        """expr_stmt : testlist_star_expr augassign yield_expr_or_testlist
                     | testlist_star_expr equals_yield_expr_or_testlist_list_opt
                     | testlist equals_yield_expr_or_testlist_list_opt
                     | test_comma_list_opt star_expr comma_test_list equals_yield_expr_or_testlist
                     | test_comma_list_opt star_expr comma_opt test_comma_list_opt equals_yield_expr_or_testlist
        """
        lenp = len(p)
        p1, p2 = p[1], p[2]
        p1 = [] if p1 is None else p1
        if isinstance(p1, ast.Tuple):
            p1 = [p1]
        for targ in p1:
            store_ctx(targ)
        if lenp == 3:
            if p2 is None and len(p1) == 1:
                load_ctx(p1[0])
                p0 = self.expr(p1[0])
            elif p2 is None:
                assert False
            else:
                list(map(store_ctx, p2[:-1]))
                p0 = ast.Assign(targets=p1 + p2[:-1],
                                value=p2[-1],
                                lineno=self.lineno,
                                col_offset=self.col)
        elif lenp == 4:
            op = self._augassign_op[p2]()
            p0 = ast.AugAssign(target=p1[0],
                               op=op,
                               value=p[3],
                               lineno=self.lineno,
                               col_offset=self.col)
        elif lenp == 5 or lenp == 6:
            if lenp == 5:
                targs, rhs = p[3], p[4][0]
            else:
                targs, rhs = (p[4] or []), p[5][0]
            store_ctx(p2)
            for targ in targs:
                store_ctx(targ)
            p1.append(p2)
            p1.extend(targs)
            p1 = [ast.Tuple(elts=p1,
                            ctx=ast.Store(),
                            lineno=self.lineno,
                            col_offset=self.col)]
            p0 = ast.Assign(targets=p1,
                            value=rhs,
                            lineno=self.lineno,
                            col_offset=self.col)
        else:
            assert False
        p[0] = p0

    def p_test_comma(self, p):
        """test_comma : test COMMA"""
        p[0] = [p[1]]

    def p_comma_opt(self, p):
        """comma_opt : COMMA
                     | empty
        """
        if len(p) == 2:
            p[0] = p[1]

    def p_test_or_star_expr(self, p):
        """test_or_star_expr : test
                             | star_expr
        """
        p[0] = p[1]

    def p_comma_test_or_star_expr(self, p):
        """comma_test_or_star_expr : COMMA test_or_star_expr"""
        p[0] = [p[2]]

    def p_testlist_star_expr(self, p):
        """testlist_star_expr : test_or_star_expr comma_test_or_star_expr_list comma_opt
                              | test_or_star_expr comma_opt
        """
        p1, p2 = p[1], p[2]
        if p2 is None:
            p0 = [p1]
        elif p2 == ',':
            p0 = [ast.Tuple(elts=[p1],
                            ctx=ast.Load(),
                            lineno=self.lineno,
                            col_offset=self.col)]
        else:
            p0 = [ast.Tuple(elts=[p1] + p2,
                            ctx=ast.Load(),
                            lineno=self.lineno,
                            col_offset=self.col)]
        p[0] = p0

    def p_augassign(self, p):
        """augassign : PLUSEQUAL
                     | MINUSEQUAL
                     | TIMESEQUAL
                     | DIVEQUAL
                     | MODEQUAL
                     | AMPERSANDEQUAL
                     | PIPEEQUAL
                     | XOREQUAL
                     | LSHIFTEQUAL
                     | RSHIFTEQUAL
                     | POWEQUAL
                     | DOUBLEDIVEQUAL
        """
        p[0] = p[1]

    def p_yield_expr_or_testlist(self, p):
        """yield_expr_or_testlist : yield_expr
                                  | testlist
        """
        p[0] = p[1]

    def p_equals_yield_expr_or_testlist(self, p):
        """equals_yield_expr_or_testlist : EQUALS yield_expr_or_testlist"""
        p[0] = [p[2]]

    #
    # For normal assignments, additional restrictions enforced
    # by the interpreter
    #
    def p_del_stmt(self, p):
        """del_stmt : DEL exprlist"""
        p2 = p[2]
        for targ in p2:
            targ.ctx = ast.Del()
        p0 = ast.Delete(targets=p2, lineno=self.lineno, col_offset=self.col)
        p[0] = p0

    def p_pass_stmt(self, p):
        """pass_stmt : PASS"""
        p[0] = ast.Pass(lineno=self.lineno, col_offset=self.col)

    def p_flow_stmt(self, p):
        """flow_stmt : break_stmt
                     | continue_stmt
                     | return_stmt
                     | raise_stmt
                     | yield_stmt
        """
        p[0] = p[1]

    def p_break_stmt(self, p):
        """break_stmt : BREAK"""
        p[0] = ast.Break(lineno=self.lineno, col_offset=self.col)

    def p_continue_stmt(self, p):
        """continue_stmt : CONTINUE"""
        p[0] = ast.Continue(lineno=self.lineno, col_offset=self.col)

    def p_return_stmt(self, p):
        """return_stmt : RETURN testlist_opt"""
        p[0] = ast.Return(value=p[2], lineno=self.lineno, col_offset=self.col)

    def p_yield_stmt(self, p):
        """yield_stmt : yield_expr"""
        p[0] = self.expr(p[1])

    def p_raise_stmt(self, p):
        """raise_stmt : RAISE
                      | RAISE test
                      | RAISE test FROM test
        """
        lenp = len(p)
        cause = None
        if lenp == 2:
            exc = None
        elif lenp == 3:
            exc = p[2]
        elif lenp == 5:
            exc = p[2]
            cause = p[4]
        else:
            assert False
        p0 = ast.Raise(exc=exc,
                       cause=cause,
                       lineno=self.lineno,
                       col_offset=self.col)
        p[0] = p0

    def p_import_stmt(self, p):
        """import_stmt : import_name
                       | import_from
        """
        p[0] = p[1]

    def p_import_name(self, p):
        """import_name : IMPORT dotted_as_names
        """
        p[0] = ast.Import(names=p[2], lineno=self.lineno, col_offset=self.col)

    def p_import_from_pre(self, p):
        """import_from_pre : FROM period_or_ellipsis_list_opt dotted_name
                           | FROM period_or_ellipsis_list
        """
        if len(p) == 3:
            p0 = p[2]
        elif len(p) == 4:
            p2, p3 = p[2], p[3]
            p0 = p3 if p2 is None else p2 + p3
        else:
            assert False
        p[0] = p0

    def p_import_from_post(self, p):
        """import_from_post : TIMES
                            | LPAREN import_as_names RPAREN
                            | import_as_names
        """
        if len(p) == 2:
            if p[1] == '*':
                p[1] = [ast.alias(name='*', asname=None)]
            p0 = p[1]
        elif len(p) == 4:
            p0 = p[2]
        else:
            assert False
        p[0] = p0

    def p_import_from(self, p):
        """import_from : import_from_pre IMPORT import_from_post
        """
        # note below: the ('.' | '...') is necessary because '...' is
        # tokenized as ELLIPSIS
        p1 = p[1]
        mod = p1.lstrip('.')
        lvl = len(p1) - len(mod)
        mod = mod or None
        p[0] = ast.ImportFrom(module=mod,
                              names=p[3],
                              level=lvl,
                              lineno=self.lineno,
                              col_offset=self.col)

    def p_period_or_ellipsis(self, p):
        """period_or_ellipsis : PERIOD
                              | ELLIPSIS
        """
        p[0] = p[1]

    def p_as_name(self, p):
        """as_name : AS NAME"""
        p[0] = p[2]

    def p_import_as_name(self, p):
        """import_as_name : NAME as_name_opt"""
        p[0] = ast.alias(name=p[1], asname=p[2])

    def p_comma_import_as_name(self, p):
        """comma_import_as_name : COMMA import_as_name
        """
        p[0] = [p[2]]

    def p_dotted_as_name(self, p):
        """dotted_as_name : dotted_name as_name_opt"""
        p0 = ast.alias(name=p[1], asname=p[2])
        p[0] = p0

    def p_comma_dotted_as_name(self, p):
        """comma_dotted_as_name : COMMA dotted_as_name"""
        p[0] = [p[2]]

    def p_import_as_names(self, p):
        """import_as_names : import_as_name comma_import_as_name_list_opt comma_opt
        """
        p1, p2 = p[1], p[2]
        p0 = [p1]
        if p2 is not None:
            p0.extend(p2)
        p[0] = p0

    def p_dotted_as_names(self, p):
        """dotted_as_names : dotted_as_name comma_dotted_as_name_list_opt"""
        p1, p2 = p[1], p[2]
        p0 = [p1]
        if p2 is not None:
            p0.extend(p2)
        p[0] = p0

    def p_period_name(self, p):
        """period_name : PERIOD NAME"""
        p[0] = p[1] + p[2]

    def p_dotted_name(self, p):
        """dotted_name : NAME
                       | NAME period_name_list
        """
        p[0] = p[1] if len(p) == 2 else p[1] + p[2]

    def p_comma_name(self, p):
        """comma_name : COMMA NAME"""
        p[0] = [p[2]]

    def p_global_stmt(self, p):
        """global_stmt : GLOBAL NAME comma_name_list_opt"""
        p2, p3 = p[2], p[3]
        names = [p2]
        if p3 is not None:
            names += p3
        p[0] = ast.Global(names=names, lineno=self.lineno, col_offset=self.col)

    def p_nonlocal_stmt(self, p):
        """nonlocal_stmt : NONLOCAL NAME comma_name_list_opt"""
        p2, p3 = p[2], p[3]
        names = [p2]
        if p3 is not None:
            names += p3
        p[0] = ast.Nonlocal(names=names,
                            lineno=self.lineno,
                            col_offset=self.col)

    def p_comma_test(self, p):
        """comma_test : COMMA test"""
        p[0] = [p[2]]

    def p_assert_stmt(self, p):
        """assert_stmt : ASSERT test comma_test_opt"""
        p2, p3 = p[2], p[3]
        if p3 is not None:
            if len(p3) != 1:
                assert False
            p3 = p3[0]
        p0 = ast.Assert(test=p2,
                        msg=p3,
                        lineno=self.lineno,
                        col_offset=self.col)
        p[0] = p0

    def p_compound_stmt(self, p):
        """compound_stmt : if_stmt
                         | while_stmt
                         | for_stmt
                         | try_stmt
                         | with_stmt
                         | funcdef
                         | classdef
                         | decorated
        """
        p[0] = p[1]

    def p_elif_part(self, p):
        """elif_part : ELIF test COLON suite"""
        p[0] = [ast.If(test=p[2],
                       body=p[4],
                       orelse=[],
                       lineno=self.lineno,
                       col_offset=self.col)]

    def p_else_part(self, p):
        """else_part : ELSE COLON suite"""
        p[0] = p[3]

    def p_if_stmt(self, p):
        """if_stmt : IF test COLON suite elif_part_list_opt
                   | IF test COLON suite elif_part_list_opt else_part
        """
        lastif = ast.If(test=p[2],
                        body=p[4],
                        orelse=[],
                        lineno=self.lineno,
                        col_offset=self.col)
        p0 = [lastif]
        p5 = p[5]
        p6 = p[6] if len(p) > 6 else []
        if p5 is not None:
            for elseif in p5:
                lastif.orelse.append(elseif)
                lastif = elseif
        lastif.orelse = p6
        p[0] = p0

    def p_while_stmt(self, p):
        """while_stmt : WHILE test COLON suite
                      | WHILE test COLON suite else_part
        """
        p5 = p[5] if len(p) > 5 else []
        p[0] = [ast.While(test=p[2],
                          body=p[4],
                          orelse=p5,
                          lineno=self.lineno,
                          col_offset=self.col)]

    def p_for_stmt(self, p):
        """for_stmt : FOR exprlist IN testlist COLON suite
                    | FOR exprlist IN testlist COLON suite else_part
        """
        p2 = p[2]
        p7 = p[7] if len(p) > 7 else []
        if len(p2) == 1:
            p2 = p2[0]
            store_ctx(p2)
        else:
            for x in p2:
                store_ctx(x)
            p2 = ast.Tuple(elts=p2,
                           ctx=ast.Store(),
                           lineno=self.lineno,
                           col_offset=self.col)
        p[0] = [ast.For(target=p2,
                        iter=p[4],
                        body=p[6],
                        orelse=p7,
                        lineno=self.lineno,
                        col_offset=self.col)]

    def p_except_part(self, p):
        """except_part : except_clause COLON suite"""
        p0 = p[1]
        p0.body = p[3]
        p[0] = [p0]

    def p_finally_part(self, p):
        """finally_part : FINALLY COLON suite"""
        p[0] = p[3]

    def p_try_stmt(self, p):
        """try_stmt : TRY COLON suite except_part_list else_part finally_part_opt
                    | TRY COLON suite except_part_list finally_part_opt
                    | TRY COLON suite finally_part
        """
        lenp = len(p)
        t = ast.Try(body=p[3], lineno=self.lineno, col_offset=self.col)
        if lenp == 7:
            p5, p6 = p[5], p[6]
            t.handlers = p[4]
            t.orelse = [] if p5 is None else p5
            t.finalbody = [] if p6 is None else p6
        elif lenp == 6:
            p5 = p[5]
            t.handlers = p[4]
            t.orelse = []
            t.finalbody = [] if p5 is None else p5
        else:
            t.handlers = []
            t.orelse = []
            t.finalbody = p[4]
        p[0] = [t]

    def p_with_stmt(self, p):
        """with_stmt : WITH with_item COLON suite
                     | WITH with_item comma_with_item_list COLON suite
        """
        p2, p3 = [p[2]], p[3]
        if len(p) == 5:
            body = p[4]
        else:
            p2 += p3
            body = p[5]
        p[0] = [ast.With(items=p2,
                         body=body,
                         lineno=self.lineno,
                         col_offset=self.col)]

    def p_as_expr(self, p):
        """as_expr : AS expr"""
        p2 = p[2]
        store_ctx(p2)
        p[0] = p2

    def p_with_item(self, p):
        """with_item : test
                     | test as_expr
        """
        p2 = p[2] if len(p) > 2 else None
        p[0] = ast.withitem(context_expr=p[1], optional_vars=p2)

    def p_comma_with_item(self, p):
        """comma_with_item : COMMA with_item"""
        p[0] = [p[2]]

    def p_except_clause(self, p):
        """except_clause : EXCEPT
                         | EXCEPT test as_name_opt
        """
        if len(p) == 2:
            p0 = ast.ExceptHandler(type=None,
                                   name=None,
                                   lineno=self.lineno,
                                   col_offset=self.col)
        else:
            p0 = ast.ExceptHandler(type=p[2],
                                   name=p[3],
                                   lineno=self.lineno,
                                   col_offset=self.col)
        p[0] = p0

    def p_suite(self, p):
        """suite : simple_stmt
                 | NEWLINE INDENT stmt_list DEDENT
        """
        p[0] = p[1] if len(p) == 2 else p[3]

    def p_test(self, p):
        """test : or_test
                | or_test IF or_test ELSE test
                | lambdef
        """
        if len(p) == 2:
            p0 = p[1]
        else:
            p0 = ast.IfExp(test=p[3],
                           body=p[1],
                           orelse=p[5],
                           lineno=self.lineno,
                           col_offset=self.col)
        p[0] = p0

    def p_test_nocond(self, p):
        """test_nocond : or_test
                       | lambdef_nocond
        """
        p[0] = p[1]

    def p_lambdef(self, p):
        """lambdef : LAMBDA varargslist_opt COLON test"""
        p2, p4 = p[2], p[4]
        if p2 is None:
            args = ast.arguments(args=[],
                                 vararg=None,
                                 kwonlyargs=[],
                                 kw_defaults=[],
                                 kwarg=None,
                                 defaults=[])
        else:
            args = p2
        p0 = ast.Lambda(args=args,
                        body=p4,
                        lineno=self.lineno,
                        col_offset=self.col)
        p[0] = p0

    def p_lambdef_nocond(self, p):
        """lambdef_nocond : LAMBDA varargslist_opt COLON test_nocond"""
        assert False

    def p_or_test(self, p):
        """or_test : and_test or_and_test_list_opt"""
        p2 = p[2]
        if p2 is None:
            p0 = p[1]
        elif len(p2) == 2:
            p0 = ast.BoolOp(op=p2[0],
                            values=[p[1], p2[1]],
                            lineno=self.lineno,
                            col_offset=self.col)
        else:
            p0 = ast.BoolOp(op=p2[0],
                            values=[p[1]] + p2[1::2],
                            lineno=self.lineno,
                            col_offset=self.col)
        p[0] = p0

    def p_or_and_test(self, p):
        """or_and_test : OR and_test"""
        p[0] = [ast.Or(), p[2]]

    def p_and_test(self, p):
        """and_test : not_test and_not_test_list_opt"""
        p2 = p[2]
        if p2 is None:
            p0 = p[1]
        elif len(p2) == 2:
            p0 = ast.BoolOp(op=p2[0],
                            values=[p[1], p2[1]],
                            lineno=self.lineno,
                            col_offset=self.col)
        else:
            p0 = ast.BoolOp(op=p2[0],
                            values=[p[1]] + p2[1::2],
                            lineno=self.lineno,
                            col_offset=self.col)
        p[0] = p0

    def p_and_not_test(self, p):
        """and_not_test : AND not_test"""
        p[0] = [ast.And(), p[2]]

    def p_not_test(self, p):
        """not_test : NOT not_test
                    | comparison
        """
        if len(p) == 2:
            p0 = p[1]
        else:
            p0 = ast.UnaryOp(op=ast.Not(),
                             operand=p[2],
                             lineno=self.lineno,
                             col_offset=self.col)
        p[0] = p0

    def p_comparison(self, p):
        """comparison : expr comp_op_expr_list_opt"""
        p2 = p[2]
        if p2 is None:
            p0 = p[1]
        else:
            p0 = ast.Compare(left=p[1],
                             ops=p2[::2],
                             comparators=p2[1::2],
                             lineno=self.lineno,
                             col_offset=self.col)
        p[0] = p0

    def p_comp_op_expr(self, p):
        """comp_op_expr : comp_op expr"""
        p[0] = [p[1], p[2]]

    _comp_ops = {
        '<': ast.Lt,
        '>': ast.Gt,
        '==': ast.Eq,
        '>=': ast.GtE,
        '<=': ast.LtE,
        '!=': ast.NotEq,
        'in': ast.In,
        ('not', 'in'): ast.NotIn,
        'is': ast.Is,
        ('is', 'not'): ast.IsNot
    }

    def p_comp_op(self, p):
        """comp_op : LT
                   | GT
                   | EQ
                   | GE
                   | LE
                   | NE
                   | IN
                   | NOT IN
                   | IS
                   | IS NOT
        """
        key = p[1] if len(p) == 2 else (p[1], p[2])
        p[0] = self._comp_ops[key]()

    def p_star_expr(self, p):
        """star_expr : TIMES expr"""
        p[0] = ast.Starred(value=p[2],
                           ctx=ast.Load(),
                           lineno=self.lineno,
                           col_offset=self.col)

    def _binop_combine(self, p1, p2):
        """Combines binary operations"""
        if p2 is None:
            p0 = p1
        elif isinstance(p2, ast.BinOp):
            p2.left = p1
            p0 = p2
        elif isinstance(p2, Sequence) and isinstance(p2[0], ast.BinOp):
            p0 = p2[0]
            p0.left = p1
            for bop in p2[1:]:
                bop.left = p0
                p0 = bop
        else:
            p0 = p1 + p2
        return p0

    def p_expr(self, p):
        """expr : xor_expr
                | xor_expr pipe_xor_expr_list
        """
        p[0] = self._binop_combine(p[1], p[2] if len(p) > 2 else None)

    def p_pipe_xor_expr(self, p):
        """pipe_xor_expr : PIPE xor_expr"""
        p[0] = [ast.BinOp(left=None,
                          op=ast.BitOr(),
                          right=p[2],
                          lineno=self.lineno,
                          col_offset=self.col)]

    def p_xor_expr(self, p):
        """xor_expr : and_expr xor_and_expr_list_opt"""
        p[0] = self._binop_combine(p[1], p[2])

    def p_xor_and_expr(self, p):
        """xor_and_expr : XOR and_expr"""
        p[0] = [ast.BinOp(left=None,
                          op=ast.BitXor(),
                          right=p[2],
                          lineno=self.lineno,
                          col_offset=self.col)]

    def p_and_expr(self, p):
        """and_expr : shift_expr ampersand_shift_expr_list_opt"""
        p[0] = self._binop_combine(p[1], p[2])

    def p_ampersand_shift_expr(self, p):
        """ampersand_shift_expr : AMPERSAND shift_expr"""
        p[0] = [ast.BinOp(left=None,
                          op=ast.BitAnd(),
                          right=p[2],
                          lineno=self.lineno,
                          col_offset=self.col)]

    def p_shift_expr(self, p):
        """shift_expr : arith_expr shift_arith_expr_list_opt"""
        p[0] = self._binop_combine(p[1], p[2])

    def p_shift_arith_expr(self, p):
        """shift_arith_expr : LSHIFT arith_expr
                            | RSHIFT arith_expr
        """
        op = ast.LShift() if p[1] == '<<' else ast.RShift()
        p[0] = [ast.BinOp(left=None,
                          op=op,
                          right=p[2],
                          lineno=self.lineno,
                          col_offset=self.col)]

    def p_arith_expr(self, p):
        """arith_expr : term
                      | term pm_term_list
        """
        p2 = p[2] if len(p) > 2 else None
        if p2 is None:
            p0 = p[1]
        elif len(p2) == 2:
            p0 = ast.BinOp(left=p[1],
                           op=p2[0],
                           right=p2[1],
                           lineno=self.lineno,
                           col_offset=self.col)
        else:
            left = p[1]
            for op, right in zip(p2[::2], p2[1::2]):
                left = ast.BinOp(left=left,
                                 op=op,
                                 right=right,
                                 lineno=self.lineno,
                                 col_offset=self.col)
            p0 = left
        p[0] = p0

    _term_binops = {
        '+': ast.Add,
        '-': ast.Sub,
        '*': ast.Mult,
        '/': ast.Div,
        '%': ast.Mod,
        '//': ast.FloorDiv
    }

    def p_pm_term(self, p):
        """pm_term : PLUS term
                   | MINUS term
        """
        op = self._term_binops[p[1]]()
        p[0] = [op, p[2]]

    def p_term(self, p):
        """term : factor op_factor_list_opt"""
        p2 = p[2]
        if p2 is None:
            p0 = p[1]
        elif len(p2) == 2:
            p0 = ast.BinOp(left=p[1],
                           op=p2[0],
                           right=p2[1],
                           lineno=self.lineno,
                           col_offset=self.col)
        else:
            left = p[1]
            for op, right in zip(p2[::2], p2[1::2]):
                left = ast.BinOp(left=left,
                                 op=op,
                                 right=right,
                                 lineno=self.lineno,
                                 col_offset=self.col)
            p0 = left
        p[0] = p0

    def p_op_factor(self, p):
        """op_factor : TIMES factor
                     | DIVIDE factor
                     | MOD factor
                     | DOUBLEDIV factor
        """
        op = self._term_binops[p[1]]()
        p[0] = [op, p[2]]

    _factor_ops = {'+': ast.UAdd, '-': ast.USub, '~': ast.Invert}

    def p_factor(self, p):
        """factor : PLUS factor
                  | MINUS factor
                  | TILDE factor
                  | power
        """
        if len(p) == 2:
            p0 = p[1]
        else:
            op = self._factor_ops[p[1]]()
            p0 = ast.UnaryOp(op=op,
                             operand=p[2],
                             lineno=self.lineno,
                             col_offset=self.col)
        p[0] = p0

    def p_power(self, p):
        """power : atom trailer_list_opt
                 | atom trailer_list_opt POW factor
        """
        p1, p2 = p[1], p[2]
        p0 = leader = p1
        if p2 is None:
            p2 = []
        for trailer in p2:
            if isinstance(trailer, (ast.Index, ast.Slice)):
                p0 = ast.Subscript(value=leader,
                                   slice=trailer,
                                   ctx=ast.Load(),
                                   lineno=self.lineno,
                                   col_offset=self.col)
            elif isinstance(trailer, Mapping):
                p0 = ast.Call(func=leader,
                              lineno=self.lineno,
                              col_offset=self.col, **trailer)
            elif isinstance(trailer, str):
                if trailer == '?':
                    p0 = xonsh_help(leader, lineno=self.lineno, col=self.col)
                elif trailer == '??':
                    p0 = xonsh_superhelp(leader,
                                         lineno=self.lineno,
                                         col=self.col)
                else:
                    p0 = ast.Attribute(value=leader,
                                       attr=trailer,
                                       ctx=ast.Load(),
                                       lineno=self.lineno,
                                       col_offset=self.col)
            else:
                assert False
            leader = p0

        # actual power rule
        if len(p) == 5:
            p0 = ast.BinOp(left=p0,
                           op=ast.Pow(),
                           right=p[4],
                           lineno=self.lineno,
                           col_offset=self.col)
        p[0] = p0

    def p_yield_expr_or_testlist_comp(self, p):
        """yield_expr_or_testlist_comp : yield_expr
                                       | testlist_comp
        """
        p[0] = p[1]

    def p_atom(self, p):
        """atom : LPAREN yield_expr_or_testlist_comp_opt RPAREN
                | LBRACKET testlist_comp_opt RBRACKET
                | LBRACE dictorsetmaker_opt RBRACE
                | NAME
                | number
                | string_literal_list
                | ELLIPSIS
                | NONE
                | TRUE
                | FALSE
                | REGEXPATH
                | DOLLAR_NAME
                | DOLLAR_LBRACE test RBRACE
                | DOLLAR_LPAREN subproc RPAREN
                | DOLLAR_LBRACKET subproc RBRACKET
        """
        p1 = p[1]
        if len(p) == 2:
            # plain-old atoms
            bt = '`'
            if isinstance(p1, (ast.Num, ast.Str, ast.Bytes)):
                pass
            elif p1 == 'True':
                p1 = ast.NameConstant(value=True,
                                      lineno=self.lineno,
                                      col_offset=self.col)
            elif p1 == 'False':
                p1 = ast.NameConstant(value=False,
                                      lineno=self.lineno,
                                      col_offset=self.col)
            elif p1 == 'None':
                p1 = ast.NameConstant(value=None,
                                      lineno=self.lineno,
                                      col_offset=self.col)
            elif p1 == '...':
                p1 = ast.Ellipsis(lineno=self.lineno, col_offset=self.col)
            elif p1.startswith(bt) and p1.endswith(bt):
                p1 = ast.Str(s=p1.strip(bt),
                             lineno=self.lineno,
                             col_offset=self.col)
                p1 = xonsh_regexpath(p1, lineno=self.lineno, col=self.col)
            elif p1.startswith('$'):
                p1 = self._envvar_by_name(p1[1:],
                                          lineno=self.lineno,
                                          col=self.col)
            else:
                p1 = ast.Name(id=p1,
                              ctx=ast.Load(),
                              lineno=self.lineno,
                              col_offset=self.col)
            p[0] = p1
            return
        p2 = p[2]
        if p2 is None:
            # empty container atoms
            if p1 == '(':
                p0 = ast.Tuple(elts=[],
                               ctx=ast.Load(),
                               lineno=self.lineno,
                               col_offset=self.col)
            elif p1 == '[':
                p0 = ast.List(elts=[],
                              ctx=ast.Load(),
                              lineno=self.lineno,
                              col_offset=self.col)
            elif p1 == '{':
                p0 = ast.Dict(keys=[],
                              values=[],
                              ctx=ast.Load(),
                              lineno=self.lineno,
                              col_offset=self.col)
            else:
                assert False
        elif p1 == '(':
            # filled, possible group container tuple atoms
            if isinstance(p2, ast.AST):
                p0 = p2
                p0._real_tuple = True
            elif len(p2) == 1 and isinstance(p2[0], ast.AST):
                p0 = p2[0]
            else:
                assert False
        elif p1 == '[':
            if isinstance(p2, ast.GeneratorExp):
                p0 = ast.ListComp(elt=p2.elt,
                                  generators=p2.generators,
                                  lineno=p2.lineno,
                                  col_offset=p2.col_offset)
            else:
                if isinstance(p2, ast.Tuple):
                    if hasattr(p2, '_real_tuple') and p2._real_tuple:
                        elts = [p2]
                    else:
                        elts = p2.elts
                else:
                    elts = [p2]
                p0 = ast.List(elts=elts,
                              ctx=ast.Load(),
                              lineno=self.lineno,
                              col_offset=self.col)
        elif p1 == '{':
            p0 = p2
        elif p1.startswith('$'):
            p0 = self._dollar_rules(p)
        else:
            assert False
        p[0] = p0

    def p_string_literal(self, p):
        """string_literal : STRING"""
        s = eval(p[1])
        cls = ast.Bytes if p[1].startswith('b') else ast.Str
        p[0] = cls(s=s, lineno=self.lineno, col_offset=self.col)

    def p_string_literal_list(self, p):
        """string_literal_list : string_literal
                               | string_literal_list string_literal
        """
        if len(p) == 3:
            p[1].s += p[2].s
        p[0] = p[1]

    def p_number(self, p):
        """number : NUMBER"""
        p[0] = ast.Num(n=eval(p[1]), lineno=self.lineno, col_offset=self.col)

    def p_testlist_comp(self, p):
        """testlist_comp : test_or_star_expr comp_for
                         | test_or_star_expr comma_opt
                         | test_or_star_expr comma_test_or_star_expr_list comma_opt
        """
        p1, p2 = p[1], p[2]
        if len(p) == 3:
            if p2 is None:
                # split out grouping parentheses.
                p0 = p1
            elif p2 == ',':
                p0 = ast.Tuple(elts=[p1],
                               ctx=ast.Load(),
                               lineno=self.lineno,
                               col_offset=self.col)
            elif 'comps' in p2:
                p0 = ast.GeneratorExp(elt=p1,
                                      generators=p2['comps'],
                                      lineno=self.lineno,
                                      col_offset=self.col)
            else:
                assert False
        elif len(p) == 4:
            p0 = ast.Tuple(elts=[p1],
                           ctx=ast.Load(),
                           lineno=self.lineno,
                           col_offset=self.col)
            if p2 is not None:
                p0.elts.extend(p2)
            else:
                assert False
        else:
            assert False
        p[0] = p0

    def p_trailer(self, p):
        """trailer : LPAREN arglist_opt RPAREN
                   | LBRACKET subscriptlist RBRACKET
                   | PERIOD NAME
                   | DOUBLE_QUESTION
                   | QUESTION
        """
        p1 = p[1]
        p2 = p[2] if len(p) > 2 else None
        if p1 == '[':
            p0 = [p2]
        elif p1 == '(':
            p0 = [p2 or dict(args=[], keywords=[], starargs=None, kwargs=None)]
        elif p1 == '.':
            p0 = [p2]
        elif p1 == '?' or p1 == '??':
            p0 = [p1]
        else:
            assert False
        p[0] = p0

    def p_subscriptlist(self, p):
        """subscriptlist : subscript comma_subscript_list_opt comma_opt"""
        p1, p2 = p[1], p[2]
        if p2 is not None:
            p1.value = ast.Tuple(elts=[p1.value] + [x.value for x in p2],
                                 ctx=ast.Load(),
                                 lineno=self.lineno,
                                 col_offset=self.col)
        p[0] = p1

    def p_comma_subscript(self, p):
        """comma_subscript : COMMA subscript"""
        p[0] = [p[2]]

    def p_subscript(self, p):
        """subscript : test
                     | test_opt COLON test_opt sliceop_opt
        """
        if len(p) == 2:
            p0 = ast.Index(value=p[1])
        else:
            p0 = ast.Slice(lower=p[1], upper=p[3], step=p[4])
        p[0] = p0

    def p_sliceop(self, p):
        """sliceop : COLON test_opt"""
        p[0] = p[2]

    def p_expr_or_star_expr(self, p):
        """expr_or_star_expr : expr
                             | star_expr
        """
        p[0] = p[1]

    def p_comma_expr_or_star_expr(self, p):
        """comma_expr_or_star_expr : COMMA expr_or_star_expr"""
        p[0] = [p[2]]

    def p_exprlist(self, p):
        """exprlist : expr_or_star_expr comma_expr_or_star_expr_list comma_opt
                    | expr_or_star_expr comma_opt
        """
        p1, p2 = p[1], p[2]
        p3 = p[3] if len(p) == 4 else None
        if p2 is None and p3 is None:
            p0 = [p1]
        elif p2 == ',' and p3 is None:
            p0 = [p1]
        elif p2 is not None:
            p2.insert(0, p1)
            p0 = p2
        else:
            assert False
        p[0] = p0

    def p_testlist(self, p):
        """testlist : test comma_test_list COMMA
                    | test comma_test_list
                    | test COMMA
                    | test
        """
        lenp = len(p)
        p1 = p[1]
        if lenp > 2:
            if isinstance(p1, ast.Tuple) and \
                    (hasattr(p1, '_real_tuple') and p1._real_tuple):
                p1 = ast.Tuple(elts=[p1],
                               ctx=ast.Load(),
                               lineno=self.lineno,
                               col_offset=self.col)
            else:
                p1 = ensure_has_elts(p1,
                                     lineno=self.lineno,
                                     col_offset=self.col)
            p2 = p[2] if lenp > 2 else []
            p2 = [] if p2 == ',' else p2
            p1.elts += p2
        p[0] = p1

    def p_comma_item(self, p):
        """comma_item : COMMA test COLON test"""
        p[0] = [p[2], p[4]]

    def p_dictorsetmaker(self, p):
        """dictorsetmaker : test COLON test comp_for
                          | test COLON test comma_item_list comma_opt
                          | test COLON testlist
                          | test comp_for
                          | testlist
        """
        p1 = p[1]
        lenp = len(p)
        if lenp == 2:
            elts = [p1]
            if isinstance(p1, ast.Tuple) and \
                    not (hasattr(p1, '_real_tuple') and p1._real_tuple):
                elts = p1.elts
            p0 = ast.Set(elts=elts,
                         ctx=ast.Load(),
                         lineno=self.lineno,
                         col_offset=self.col)
        elif lenp == 3:
            comps = p[2].get('comps', [])
            p0 = ast.SetComp(elt=p1,
                             generators=comps,
                             lineno=self.lineno,
                             col_offset=self.col)
        elif lenp == 4:
            p3 = p[3]
            vals = [p3]
            if isinstance(p3, ast.Tuple) and \
                    not (hasattr(p3, '_real_tuple') and p3._real_tuple):
                vals = p3.elts
            p0 = ast.Dict(keys=[p1],
                          values=vals,
                          ctx=ast.Load(),
                          lineno=self.lineno,
                          col_offset=self.col)
        elif lenp == 5:
            comps = p[4].get('comps', [])
            p0 = ast.DictComp(key=p1,
                              value=p[3],
                              generators=comps,
                              lineno=self.lineno,
                              col_offset=self.col)
        elif lenp == 6:
            p3, p4 = p[3], p[4]
            keys = [p1] + p4[::2]
            values = [p3] + p4[1::2]
            p0 = ast.Dict(keys=keys,
                          values=values,
                          ctx=ast.Load(),
                          lineno=self.lineno,
                          col_offset=self.col)
        else:
            assert False
        p[0] = p0

    def p_classdef(self, p):
        """classdef : CLASS NAME func_call_opt COLON suite"""
        p3 = p[3]
        b, kw = ([], []) if p3 is None else (p3['args'], p3['keywords'])
        c = ast.ClassDef(name=p[2],
                         bases=b,
                         keywords=kw,
                         starargs=None,
                         kwargs=None,
                         body=p[5],
                         decorator_list=[],
                         lineno=self.lineno,
                         col_offset=self.col)
        p[0] = [c]

    def _set_arg(self, args, arg, ensure_kw=False):
        if isinstance(arg, ast.keyword):
            args['keywords'].append(arg)
        elif ensure_kw:
            raise SyntaxError('{0} must be a keyword'.format(arg))
        else:
            args['args'].append(arg)

    def p_arglist(self, p):
        """arglist : argument comma_opt
                   | argument_comma_list argument comma_opt
                   | argument_comma_list_opt TIMES test comma_argument_list_opt
                   | argument_comma_list_opt TIMES test COMMA POW test
                   | argument_comma_list_opt TIMES test comma_argument_list COMMA POW test
                   | argument_comma_list_opt POW test
        """
        lenp = len(p)
        p1, p2 = p[1], p[2]
        p0 = {'args': [], 'keywords': [], 'starargs': None, 'kwargs': None}
        if lenp == 3:
            self._set_arg(p0, p1)
        elif lenp == 4 and p2 != '**':
            for arg in p1:
                self._set_arg(p0, arg)
            self._set_arg(p0, p2)
        elif lenp == 4 and p2 == '**':
            if p1 is not None:
                for arg in p1:
                    self._set_arg(p0, arg)
            p0['kwargs'] = p[3]
        elif lenp == 5:
            p0['starargs'], p4 = p[3], p[4]
            if p1 is not None:
                for arg in p1:
                    self._set_arg(p0, arg)
            if p4 is not None:
                for arg in p4:
                    self._set_arg(p0, arg, ensure_kw=True)
        elif lenp == 7:
            p0['starargs'], p0['kwargs'] = p[3], p[6]
            if p1 is not None:
                for arg in p1:
                    self._set_arg(p0, arg)
        elif lenp == 8:
            p0['starargs'], p4, p0['kwargs'] = p[3], p[4], p[7]
            if p1 is not None:
                for arg in p1:
                    self._set_arg(p0, arg)
            for arg in p4:
                self._set_arg(p0, arg, ensure_kw=True)
        else:
            assert False
        p[0] = p0

    def p_argument_comma(self, p):
        """argument_comma : argument COMMA"""
        p[0] = [p[1]]

    def p_comma_argument(self, p):
        """comma_argument : COMMA argument """
        p[0] = [p[2]]

    def p_argument(self, p):
        """argument : test
                    | test comp_for
                    | test EQUALS test
        """
        # Really [keyword '='] test
        # The reason that keywords are test nodes instead of NAME is that using
        # NAME results in an ambiguity.
        p1 = p[1]
        lenp = len(p)
        if lenp == 2:
            p0 = p1
        elif lenp == 3:
            p0 = ast.GeneratorExp(elt=p1,
                                  generators=p[2]['comps'],
                                  lineno=self.lineno,
                                  col_offset=self.col)
        elif lenp == 4:
            p0 = ast.keyword(arg=p1.id, value=p[3])
        else:
            assert False
        p[0] = p0

    def p_comp_iter(self, p):
        """comp_iter : comp_for
                     | comp_if
        """
        p[0] = p[1]

    def p_comp_for(self, p):
        """comp_for : FOR exprlist IN or_test comp_iter_opt"""
        targs, it, p5 = p[2], p[4], p[5]
        if len(targs) != 1:
            assert False
        targ = targs[0]
        store_ctx(targ)
        comp = ast.comprehension(target=targ, iter=it, ifs=[])
        comps = [comp]
        p0 = {'comps': comps}
        if p5 is not None:
            comps += p5.get('comps', [])
            comp.ifs += p5.get('if', [])
        p[0] = p0

    def p_comp_if(self, p):
        """comp_if : IF test_nocond comp_iter_opt"""
        p2, p3 = p[2], p[3]
        p0 = {'if': [p2]}
        if p3 is not None:
            p0['comps'] = p3.get('comps', [])
        p[0] = p0

    def p_yield_expr(self, p):
        """yield_expr : YIELD yield_arg_opt"""
        p2 = p[2]
        if p2 is None:
            p0 = ast.Yield(value=p2, lineno=self.lineno, col_offset=self.col)
        elif p2['from']:
            p0 = ast.YieldFrom(value=p2['val'],
                               lineno=self.lineno,
                               col_offset=self.col)
        else:
            p0 = ast.Yield(value=p2['val'],
                           lineno=self.lineno,
                           col_offset=self.col)
        p[0] = p0

    def p_yield_arg(self, p):
        """yield_arg : FROM test
                     | testlist
        """
        if len(p) == 2:
            p0 = {'from': False, 'val': p[1]}
        else:
            p0 = {'from': True, 'val': p[2]}
        p[0] = p0

    #
    # subprocess
    #

    def _dollar_rules(self, p):
        """These handle the special xonsh $ shell atoms by looking up
        in a special __xonsh_env__ dictionary injected in the __builtin__.
        """
        lenp = len(p)
        p1, p2 = p[1], p[2]
        col = self.col
        lineno = self.lineno
        if lenp == 3:  # $NAME
            p0 = self._envvar_by_name(p2, lineno=lineno, col=col)
        elif p1 == '${':
            xenv = self._xenv(lineno=lineno, col=col)
            idx = ast.Index(value=p2)
            p0 = ast.Subscript(value=xenv,
                               slice=idx,
                               ctx=ast.Load(),
                               lineno=lineno,
                               col_offset=col)
        elif p1 == '$(':
            p0 = xonsh_call('__xonsh_subproc_captured__', p2,
                            lineno=lineno,
                            col=col)
        elif p1 == '$[':
            p0 = xonsh_call('__xonsh_subproc_uncaptured__', p2,
                            lineno=lineno,
                            col=col)
        else:
            assert False
        return p0

    def _xenv(self, lineno=lineno, col=col):
        """Creates a new xonsh env reference."""
        return ast.Name(id='__xonsh_env__',
                        ctx=ast.Load(),
                        lineno=lineno,
                        col_offset=col)

    def _envvar_by_name(self, var, lineno=None, col=None):
        """Looks up a xonsh variable by name."""
        xenv = self._xenv(lineno=lineno, col=col)
        idx = ast.Index(value=ast.Str(s=var, lineno=lineno, col_offset=col))
        return ast.Subscript(value=xenv,
                             slice=idx,
                             ctx=ast.Load(),
                             lineno=lineno,
                             col_offset=col)

    def _subproc_cliargs(self, args, lineno=None, col=None):
        """Creates an expression for subprocess CLI arguments."""
        cliargs = currlist = empty_list(lineno=lineno, col=col)
        for arg in args:
            action = arg._cliarg_action
            if action == 'append':
                if currlist is None:
                    currlist = empty_list(lineno=lineno, col=col)
                    cliargs = binop(cliargs, ast.Add(), currlist,
                                    lineno=lineno,
                                    col=col)
                currlist.elts.append(arg)
            elif action == 'extend':
                cliargs = binop(cliargs, ast.Add(), arg,
                                lineno=lineno,
                                col=col)
                currlist = None
            elif action == 'splitlines':
                sl = call_split_lines(arg, lineno=lineno, col=col)
                cliargs = binop(cliargs, ast.Add(), sl, lineno=lineno, col=col)
                currlist = None
            elif action == 'ensure_list':
                x = ensure_list_from_str_or_list(arg, lineno=lineno, col=col)
                cliargs = binop(cliargs, ast.Add(), x, lineno=lineno, col=col)
                currlist = None
            else:
                raise ValueError("action not understood: " + action)
            del arg._cliarg_action
        return cliargs

    def p_subproc_special_atom(self, p):
        """subproc_special_atom : PIPE
                                | GT
                                | RSHIFT
        """
        p[0] = p[1]

    def p_subproc_special(self, p):
        """subproc_special : subproc_special_atom
                           | WS subproc_special_atom
                           | subproc_special_atom WS
                           | WS subproc_special_atom WS
        """
        p1 = p[1]
        if len(p) > 2 and len(p1.strip()) == 0:
            p1 = p[2]
        p[0] = ast.Str(s=p1, lineno=self.lineno, col_offset=self.col)

    def p_subproc(self, p):
        """subproc : subproc_atoms
                   | subproc_atoms WS
                   | subproc AMPERSAND
                   | subproc subproc_special subproc_atoms
                   | subproc subproc_special subproc_atoms WS
        """
        lineno = self.lineno
        col = self.col
        lenp = len(p)
        p1 = p[1]
        if lenp == 2:
            p0 = [self._subproc_cliargs(p1, lineno=lineno, col=col)]
        elif p[2] == '&':
            p0 = p1 + [ast.Str(s=p[2], lineno=lineno, col_offset=col)]
        elif lenp == 3:
            p0 = [self._subproc_cliargs(p1, lineno=lineno, col=col)]
        else:
            if len(p1) > 1 and hasattr(p1[-2], 's') and p1[-2].s != '|':
                msg = 'additional redirect following non-pipe redirect'
                self._parse_error(msg, self.currloc(lineno=lineno, column=col))
            cliargs = self._subproc_cliargs(p[3], lineno=lineno, col=col)
            p0 = p1 + [p[2], cliargs]
        # return arguments list
        p[0] = p0

    def p_subproc_atoms(self, p):
        """subproc_atoms : subproc_atom
                         | subproc_atoms WS subproc_atom
        """
        p1 = p[1]
        if len(p) < 4:
            p1 = [p1]
        else:
            p1.append(p[3])
        p[0] = p1

    def p_subproc_atom(self, p):
        """subproc_atom : subproc_arg
                        | string_literal
                        | REGEXPATH
                        | DOLLAR_NAME
                        | AT_LPAREN test RPAREN
                        | DOLLAR_LBRACE test RBRACE
                        | DOLLAR_LPAREN subproc RPAREN
                        | DOLLAR_LBRACKET subproc RBRACKET
        """
        lenp = len(p)
        p1 = p[1]
        if lenp == 2:
            if isinstance(p1, str):
                p0 = ast.Str(s=p1, lineno=self.lineno, col_offset=self.col)
                bt = '`'
                if p1.startswith(bt) and p1.endswith(bt):
                    p0.s = p1.strip(bt)
                    p0 = xonsh_regexpath(p0, lineno=self.lineno, col=self.col)
                    p0._cliarg_action = 'extend'
                elif '*' in p1:
                    p0 = xonsh_call('__xonsh_glob__',
                                    args=[p0],
                                    lineno=self.lineno,
                                    col=self.col)
                    p0._cliarg_action = 'extend'
                elif p1.startswith('$'):
                    p0 = self._envvar_by_name(p1[1:],
                                              lineno=self.lineno,
                                              col=self.col)
                    p0 = xonsh_call('__xonsh_ensure_list_of_strs__', [p0],
                                    lineno=self.lineno,
                                    col=self.col)
                    p0._cliarg_action = 'extend'
                else:
                    p0._cliarg_action = 'append'
            elif isinstance(p1, ast.AST):
                p0 = p1
                p0._cliarg_action = 'append'
            else:
                assert False
        elif p1 == '@(':
            p0 = xonsh_call('__xonsh_ensure_list_of_strs__', [p[2]],
                            lineno=self.lineno,
                            col=self.col)
            p0._cliarg_action = 'extend'
        elif p1 == '${':
            xenv = self._xenv(lineno=self.lineno, col=self.col)
            idx = ast.Index(value=p[2])
            p0 = ast.Subscript(value=xenv,
                               slice=idx,
                               ctx=ast.Load(),
                               lineno=self.lineno,
                               col_offset=self.col)
            p0._cliarg_action = 'append'
        elif p1 == '$(':
            p0 = xonsh_call('__xonsh_subproc_captured__',
                            args=p[2],
                            lineno=self.lineno,
                            col=self.col)
            p0._cliarg_action = 'splitlines'
        elif p1 == '$[':
            p0 = xonsh_call('__xonsh_subproc_uncaptured__',
                            args=p[2],
                            lineno=self.lineno,
                            col=self.col)
            p0._cliarg_action = 'splitlines'
        else:
            assert False
        p[0] = p0

    def p_subproc_arg(self, p):
        """subproc_arg : subproc_arg_part
                       | subproc_arg subproc_arg_part
        """
        # This glues the string together after parsing
        p1 = p[1]
        if len(p) == 2:
            p0 = p1
        else:
            p0 = os.path.expanduser(p1 + p[2])
        p[0] = p0

    def p_subproc_arg_part(self, p):
        """subproc_arg_part : NAME
                            | TILDE
                            | PERIOD
                            | DIVIDE
                            | MINUS
                            | PLUS
                            | COLON
                            | AT
                            | EQUALS
                            | TIMES
                            | POW
                            | MOD
                            | XOR
                            | DOUBLEDIV
                            | ELLIPSIS
                            | NONE
                            | TRUE
                            | FALSE
                            | NUMBER
        """
        # Many tokens cannot be part of this list, such as $, ', ", ()
        # Use a string atom instead.
        p[0] = p[1]

    #
    # Helpers
    #

    def p_empty(self, p):
        'empty : '
        p[0] = None

    def p_error(self, p):
        if p is None:
            self._parse_error('no further code', None)
        elif p.type == 'ERRORTOKEN':
            if isinstance(p.value, BaseException):
                raise p.value
            else:
                self._parse_error(p.value, self.currloc(lineno=p.lineno,
                                                        column=p.lexpos))
        else:
            msg = 'code: {0}'.format(p.value),
            self._parse_error(msg, self.currloc(lineno=p.lineno,
                                                column=p.lexpos))
